"""Jac Blue pass for Jaseci Ast."""
import jaclang.jac.absyntree as ast
from jaclang.jac.constant import Constants as Con
from jaclang.jac.constant import Tokens as Tok
from jaclang.jac.passes import Pass


class BluePygenPass(Pass):
    """Jac blue transpilation to python pass."""

    def before_pass(self) -> None:
        """Initialize pass."""
        self.indent_size = 4
        self.indent_level = 0
        self.debuginfo = {"jac_mods": []}
        self.preamble = ast.AstNode(parent=None, mod_link=None, kid=[], line=0)
        self.preamble.meta["py_code"] = "from __future__ import annotations\n"
        self.cur_arch = None  # tracks current architype during transpilation

    def enter_node(self, node: ast.AstNode) -> None:
        """Enter node."""
        if node:
            node.meta["py_code"] = ""
        return Pass.enter_node(self, node)

    def indent_str(self) -> str:
        """Return string for indent."""
        return " " * self.indent_size * self.indent_level

    def emit_ln(self, node: ast.AstNode, s: str) -> None:
        """Emit code to node."""
        self.emit(node, s.strip().strip("\n"))
        self.emit(node, f"  # {self.get_mod_index(node)} {node.line}\n")

    def emit_ln_unique(self, node: ast.AstNode, s: str) -> None:
        """Emit code to node."""
        if s not in node.meta["py_code"]:
            ilev = self.indent_level
            self.indent_level = 0
            self.emit_ln(node, s)
            self.indent_level = ilev

    def get_mod_index(self, node: ast.AstNode) -> int:
        """Get module index."""
        path = node.mod_link.mod_path if node.mod_link else None
        if not path:
            return -1
        if path not in self.debuginfo["jac_mods"]:
            self.debuginfo["jac_mods"].append(path)
        return self.debuginfo["jac_mods"].index(path)

    def emit(self, node: ast.AstNode, s: str) -> None:
        """Emit code to node."""
        node.meta["py_code"] += self.indent_str() + s.replace(
            "\n", "\n" + self.indent_str()
        )
        if "\n" in node.meta["py_code"]:
            node.meta["py_code"] = node.meta["py_code"].rstrip(" ")

    def needs_jac_import(self) -> None:
        """Check if import is needed."""
        self.emit_ln_unique(
            self.preamble, "from jaclang import jac_blue_import as __jac_import__"
        )

    def needs_enum(self) -> None:
        """Check if enum is needed."""
        self.emit_ln_unique(
            self.preamble,
            "from enum import Enum as __jac_Enum__, auto as __jac_auto__",
        )

    def emit_jac_error_handler(self, node: ast.AstNode) -> None:
        """Emit error handler."""
        self.emit_ln_unique(self.preamble, "import traceback as __jac_traceback__")
        self.emit_ln_unique(
            self.preamble, "from jaclang import handle_jac_error as __jac_error__"
        )
        self.emit_ln(node, "except Exception as e:")
        self.indent_level += 1
        # self.emit_ln(node, "__jac_traceback__.print_exc()")
        self.emit_ln(node, "tb = __jac_traceback__.extract_tb(e.__traceback__)")
        self.emit_ln(node, "__jac_tmp__ = __jac_error__(_jac_pycodestring_, e, tb)")
        # self.emit_ln(node, "print(__jac_tmp__)\nraise e")
        self.emit_ln(
            node,
            "e = type(e)(str(e) + '\\nOriginal Snippet:\\n' + __jac_tmp__) "
            "if 'Original Snippet:' not in str(e) else e",
        )
        self.emit_ln(node, "raise e")
        self.indent_level -= 1

    def decl_def_missing(self, decl: str = "this") -> None:
        """Warn about declaration."""
        self.error(
            f"Unable to find definition for {decl} declaration. Perhaps there's an `include` missing?"
        )

    def ds_feature_warn(self) -> None:
        """Warn about feature."""
        self.warning("Data spatial features not supported in bootstrap Jac.")

    def exit_parse(self, node: ast.Parse) -> None:
        """Sub objects.

        name: str,
        """
        self.error(f"Parse node should not be in this AST!! {node.name}")
        raise ValueError("Parse node should not be in AST after being Built!!")

    def exit_token(self, node: ast.Token) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        """
        self.emit(node, node.value)

    def exit_name(self, node: ast.Name) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        already_declared: bool,
        """
        self.emit(node, node.value)

    def exit_constant(self, node: ast.Constant) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        typ: type,
        """
        self.emit(node, node.value)

    def exit_module(self, node: ast.Module) -> None:
        """Sub objects.

        name: str,
        doc: Token,
        body: "Elements",
        """
        self.emit_ln(node, node.doc.value)
        self.emit(node, self.preamble.meta["py_code"])
        if node.body:
            self.emit(node, node.body.meta["py_code"])
        self.emit(node, f'r""" {Con.JAC_DEBUG_SPLITTER}\n')
        for i in self.debuginfo["jac_mods"]:
            self.emit(node, f"{i}\n")
        self.emit(node, f'{Con.JAC_DEBUG_SPLITTER} """\n')
        self.ir = node
        self.ir.meta["py_code"] = self.ir.meta["py_code"].rstrip()

    def exit_elements(self, node: ast.Elements) -> None:
        """Sub objects.

        elements: list[GlobalVars | Test | ModuleCode | Import | Architype | Ability | AbilitySpec],
        """
        for i in node.elements:
            self.emit(node, i.meta["py_code"])
            self.emit(node, "\n")

    def exit_global_vars(self, node: ast.GlobalVars) -> None:
        """Sub objects.

        doc: Optional["Token"],
        access: Optional[Token],
        assignments: "AssignmentList",
        is_frozen: bool,
        """
        if node.doc:
            self.emit_ln(node, node.doc.value)
        self.emit(node, node.assignments.meta["py_code"])

    # NOTE: Incomplete for Jac Purple and Red
    def exit_test(self, node: ast.Test) -> None:
        """Sub objects.

        name: Token,
        doc: Optional["Token"],
        description: Token,
        body: "CodeBlock",
        """
        self.warning("Test feature not supported in bootstrap Jac.")

    def exit_module_code(self, node: ast.ModuleCode) -> None:
        """Sub objects.

        doc: Optional["Token"],
        body: "CodeBlock",
        """
        if node.doc:
            self.emit_ln(node, node.doc.value)
        self.emit(node, node.body.meta["py_code"])

    def exit_import(self, node: ast.Import) -> None:
        """Sub objects.

        lang: Token,
        path: "ModulePath",
        alias: Optional[Token],
        items: Optional["ModuleItems"],
        is_absorb: bool,  # For includes
        self.sub_module = None
        """
        if node.lang.value == Con.JAC_LANG_IMP:  # injects module into sys.modules
            self.needs_jac_import()
            self.emit_ln(
                node,
                f"__jac_import__(target='{node.path.meta['py_code']}', base_path=__file__)",
            )
        if node.is_absorb:
            self.emit_ln(
                node,
                f"from {node.path.meta['py_code']} import *",
            )
            if node.items:
                self.warning(
                    "Includes import * in target module into current namespace."
                )
        if not node.items:
            if not node.alias:
                self.emit_ln(node, f"import {node.path.meta['py_code']}")
            else:
                self.emit_ln(
                    node,
                    f"import {node.path.meta['py_code']} as {node.alias.meta['py_code']}",
                )
        else:
            self.emit_ln(
                node,
                f"from {node.path.meta['py_code']} import {node.items.meta['py_code']}",
            )

    def exit_module_path(self, node: ast.ModulePath) -> None:
        """Sub objects.

        path: list[Token],
        """
        self.emit(node, "".join([i.value for i in node.path]))

    def exit_module_items(self, node: ast.ModuleItems) -> None:
        """Sub objects.

        items: list["ModuleItem"],
        """
        self.emit(node, ", ".join([i.meta["py_code"] for i in node.items]))

    def exit_module_item(self, node: ast.ModuleItem) -> None:
        """Sub objects.

        name: Token,
        alias: Optional[Token],
        """
        if node.alias:
            self.emit(node, node.name.value + " as " + node.alias.value)
        else:
            self.emit(node, node.name.value)

    # NOTE: Incomplete for Jac Purple and Red
    def exit_architype(self, node: ast.Architype) -> None:
        """Sub objects.

        name: Name,
        arch_type: Token,
        doc: Optional[Token],
        decorators: Optional["Decorators"],
        access: Optional[Token],
        base_classes: "BaseClasses",
        body: Optional["ArchBlock"],
        """
        if node.decorators:
            self.emit(node, node.decorators.meta["py_code"])
        if not len(node.base_classes.base_classes):
            self.emit_ln(node, f"class {node.name.meta['py_code']}:")
        else:
            self.emit_ln(
                node,
                f"class {node.name.meta['py_code']}({node.base_classes.meta['py_code']}):",
            )
        self.indent_level += 1
        if node.doc:
            self.emit_ln(node, node.doc.value)
        if node.body:
            self.emit(node, node.body.meta["py_code"])
        else:
            self.decl_def_missing(node.name.meta["py_code"])
        self.indent_level -= 1

    def exit_arch_def(self, node: ast.ArchDef) -> None:
        """Sub objects.

        doc: Optional[Token],
        mod: Optional["DottedNameList"],
        arch: "ObjectRef | NodeRef | EdgeRef | WalkerRef",
        body: "ArchBlock",
        """

    def exit_decorators(self, node: ast.Decorators) -> None:
        """Sub objects.

        calls: list["ExprType"],
        """
        for i in node.calls:
            self.emit_ln(node, "@" + i.meta["py_code"])

    def exit_base_classes(self, node: ast.BaseClasses) -> None:
        """Sub objects.

        base_classes: list[DottedNameList],
        """
        self.emit(node, ", ".join([i.meta["py_code"] for i in node.base_classes]))

    # NOTE: Incomplete for Jac Purple and Red
    def exit_ability(self, node: ast.Ability) -> None:
        """Sub objects.

        name: Name,
        is_func: bool,
        is_async: bool,
        is_static: bool,
        doc: Optional[Token],
        decorators: Optional["Decorators"],
        access: Optional[Token],
        signature: Optional["FuncSignature | TypeSpec | EventSignature"],
        body: Optional["CodeBlock"],
        arch_attached: Optional["ArchBlock"] = None,
        """
        ability_name = node.py_resolve_name()
        if node.arch_attached and ability_name == "__init__":
            return
        if node.decorators:
            self.emit(node, node.decorators.meta["py_code"])
        if isinstance(node.signature, (ast.FuncSignature, ast.EventSignature)):
            if node.arch_attached and not node.is_static:
                self.emit_ln(
                    node, f"def {ability_name}(self{node.signature.meta['py_code']}:"
                )
            else:
                if node.arch_attached and node.is_static:
                    self.emit_ln(node, "@classmethod")
                self.emit_ln(
                    node, f"def {ability_name}({node.signature.meta['py_code']}:"
                )
        else:
            if node.arch_attached:
                self.emit_ln(node, f"def {ability_name}(self):")
            else:
                self.emit_ln(node, f"def {ability_name}():")
        self.indent_level += 1
        if node.doc:
            self.emit_ln(node, node.doc.value)
        if node.body:
            self.emit_ln(node, "try:")
            self.indent_level += 1
            self.emit(node, node.body.meta["py_code"])
            self.indent_level -= 1
            self.emit_jac_error_handler(node)
        else:
            self.decl_def_missing(ability_name)
        self.indent_level -= 1

    def exit_ability_def(self, node: ast.AbilityDef) -> None:
        """Sub objects.

        doc: Optional[Token],
        mod: Optional["DottedNameList"],
        ability: AbilityRef,
        body: CodeBlock,
        """

    def exit_arch_block(self, node: ast.ArchBlock) -> None:
        """Sub objects.

        members: list["ArchHas | Ability"],
        """
        has_members = [
            i for i in node.members if isinstance(i, ast.ArchHas) and not i.is_static
        ]
        if has_members:
            self.emit_ln(node, "def __jac_has(self):")
            self.indent_level += 1
            for i in has_members:
                self.emit_ln(node, f"self._jac_has_{i.h_id}()")
            self.indent_level -= 1

        init_func = None
        for i in node.members:
            if isinstance(i, ast.Ability) and i.py_resolve_name() == "__init__":
                init_func = i
                break
        if init_func and init_func.signature:
            if init_func.decorators:
                self.emit(node, init_func.decorators.meta["py_code"])
            self.emit_ln(
                node, f"def __init__(self{init_func.signature.meta['py_code']}:"
            )
        else:
            self.emit_ln(node, "def __init__(self, *args, **kwargs):")
        self.indent_level += 1
        if not init_func:
            self.emit_ln(node, "super().__init__(*args, **kwargs)")
        if has_members:
            self.emit_ln(node, "self.__jac_has()")
        if init_func and init_func.body:
            self.emit(node, f"{init_func.body.meta['py_code']}")
        self.indent_level -= 1
        for i in node.members:
            self.emit(node, i.meta["py_code"])
            self.emit(node, "\n")

    def exit_arch_has(self, node: ast.ArchHas) -> None:
        """Sub objects.

        doc: Optional[Token],
        is_static: bool,
        access: Optional[Token],
        vars: "HasVarList",
        is_frozen: bool,
        """
        if node.is_static:
            if node.doc:
                self.emit_ln(node, node.doc.value)
            self.emit_ln(node, node.vars.meta["py_code"].replace("self.", ""))
        else:
            self.emit_ln(node, f"def _jac_has_{node.h_id}(self):")
            self.indent_level += 1
            if node.doc:
                self.emit_ln(node, node.doc.value)
            self.emit(node, node.vars.meta["py_code"])
            self.indent_level -= 1

    def exit_has_var_list(self, node: ast.HasVarList) -> None:
        """Sub objects.

        vars: list[HasVar],
        """
        for i in node.vars:
            self.emit_ln(node, i.meta["py_code"])

    def exit_has_var(self, node: ast.HasVar) -> None:
        """Sub objects.

        name: Token,
        type_tag: TypeSpec,
        value: Optional["ExprType"],
        """
        if node.value:
            self.emit(
                node,
                f"self.{node.name.value}: {node.type_tag.meta['py_code']} = {node.value.meta['py_code']}",
            )
        else:
            self.emit(
                node, f"self.{node.name.value}: {node.type_tag.meta['py_code']} = None"
            )

    def exit_type_spec_list(self, node: ast.TypeSpecList) -> None:
        """Sub objects.

        types: list[TypeSpec],
        """
        self.emit(node, "|".join([i.meta["py_code"] for i in node.types]))

    def exit_type_spec(self, node: ast.TypeSpec) -> None:
        """Sub objects.

        typ: "Token | DottedNameList",
        list_nest: TypeSpec,
        dict_nest: TypeSpec,
        """
        if node.dict_nest:
            self.emit(
                node,
                f"dict[{node.list_nest.meta['py_code']}, {node.dict_nest.meta['py_code']}]",
            )
        elif node.list_nest:
            self.emit(node, f"list[{node.list_nest.meta['py_code']}]")
        else:
            self.emit(node, node.spec_type.meta["py_code"])

    # NOTE: Incomplete for Jac Purple and Red
    def exit_event_signature(self, node: ast.EventSignature) -> None:
        """Sub objects.

        event: Token,
        arch_tag_info: Optional["TypeList | TypeSpec"],
        return_type: Optional["TypeSpec"],
        """
        self.error("Event style abilities not supported in bootstrap Jac")

    def exit_dotted_name_list(self, node: ast.DottedNameList) -> None:
        """Sub objects.

        names: list[all_refs],
        """
        self.emit(node, ".".join([i.meta["py_code"] for i in node.names]))

    def exit_func_signature(self, node: ast.FuncSignature) -> None:
        """Sub objects.

        params: Optional[FuncParams],
        return_type: Optional[TypeSpec],
        self.is_arch_attached = False
        """
        if node.params:
            if (
                isinstance(node.parent, ast.Ability)
                and node.parent.arch_attached
                and not node.parent.is_static
            ):
                self.emit(node, ", ")
            self.emit(node, node.params.meta["py_code"])
        if (
            isinstance(node.parent, ast.Ability)
            and node.parent.arch_attached
            and node.parent.py_resolve_name() == "__init__"
        ):
            self.emit(node, ", *args, **kwargs")
        self.emit(node, ")")
        if node.return_type:
            self.emit(node, f" -> {node.return_type.meta['py_code']}")

    def exit_func_params(self, node: ast.FuncParams) -> None:
        """Sub objects.

        params: list["ParamVar"],
        """
        first_out = False
        for i in node.params:
            self.emit(node, ", ") if first_out else None
            self.emit(node, i.meta["py_code"])
            first_out = True

    def exit_param_var(self, node: ast.ParamVar) -> None:
        """Sub objects.

        name: Token,
        unpack: Optional[Token],
        type_tag: "TypeSpec",
        value: Optional["ExprType"],
        """
        if node.unpack:
            self.emit(node, f"{node.unpack.value}")
        if node.value:
            self.emit(
                node,
                f"{node.name.value}: {node.type_tag.meta['py_code']} = {node.value.meta['py_code']}",
            )
        else:
            self.emit(node, f"{node.name.value}: {node.type_tag.meta['py_code']}")

    def exit_enum(self, node: ast.Enum) -> None:
        """Sub objects.

        name: Name,
        doc: Optional[Token],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[EnumBlock],
        """
        if node.decorators:
            self.emit_ln(node, node.decorators.meta["py_code"])
        if len(node.base_classes.base_classes):
            self.emit_ln(
                node,
                f"class {node.name.meta['py_code']}({node.base_classes.meta['py_code']}):",
            )
        else:
            self.needs_enum()
            self.emit_ln(node, f"class {node.name.value}(__jac_Enum__):")
        self.indent_level += 1
        if node.doc:
            self.emit_ln(node, node.doc.value)
        if node.body:
            self.emit(node, node.body.meta["py_code"])
        else:
            self.decl_def_missing(node.name.meta["py_code"])
        self.indent_level -= 1

    def exit_enum_def(self, node: ast.EnumDef) -> None:
        """Sub objects.

        doc: Optional[Token],
        mod: Optional[DottedNameList],
        body: EnumBlock,
        """

    def exit_enum_block(self, node: ast.EnumBlock) -> None:
        """Sub objects.

        stmts: list['Name|Assignment'],
        """
        for i in node.stmts:
            if isinstance(i, ast.Name):
                self.emit_ln(node, i.meta["py_code"] + " = __jac_auto__()")
            else:
                self.emit(node, i.meta["py_code"])

    def exit_code_block(self, node: ast.CodeBlock) -> None:
        """Sub objects.

        stmts: list["StmtType"],
        """
        if len(node.stmts) == 0:
            self.emit_ln(node, "pass")
        for i in node.stmts:
            self.emit(node, i.meta["py_code"])
            if len(i.meta["py_code"]) and i.meta["py_code"][-1] != "\n":
                self.emit_ln(node, "\n")

    def exit_if_stmt(self, node: ast.IfStmt) -> None:
        """Sub objects.

        condition: ExprType,
        body: CodeBlock,
        elseifs: Optional[ElseIfs],
        else_body: Optional[ElseStmt],
        """
        self.emit_ln(node, f"if {node.condition.meta['py_code']}:")
        self.indent_level += 1
        self.emit(node, node.body.meta["py_code"])
        self.indent_level -= 1
        self.emit(node, "\n")
        if node.elseifs:
            self.emit(node, node.elseifs.meta["py_code"])
        if node.else_body:
            self.emit(node, node.else_body.meta["py_code"])

    def exit_typed_ctx_block(self, node: ast.TypedCtxBlock) -> None:
        """Sub objects.

        type_ctx: TypeList,
        body: CodeBlock,
        """
        self.ds_feature_warn()

    def exit_else_ifs(self, node: ast.ElseIfs) -> None:
        """Sub objects.

        elseifs: list[IfStmt],
        """
        for i in node.elseifs:
            self.emit_ln(node, f"elif {i.condition.meta['py_code']}:")
            self.indent_level += 1
            self.emit(node, i.body.meta["py_code"])
            self.indent_level -= 1
            self.emit(node, "\n")

    def exit_else_stmt(self, node: ast.ElseStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        """
        self.emit_ln(node, "else:")
        self.indent_level += 1
        self.emit(node, node.body.meta["py_code"])
        self.indent_level -= 1
        self.emit(node, "\n")

    def exit_try_stmt(self, node: ast.TryStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        excepts: Optional[ExceptList],
        finally_body: Optional[FinallyStmt],
        """
        self.emit_ln(node, "try:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1
        if node.excepts:
            self.emit_ln(node, node.excepts.meta["py_code"])
        if node.finally_body:
            self.emit_ln(node, node.finally_body.meta["py_code"])

    def exit_except(self, node: ast.Except) -> None:
        """Sub objects.

        typ: ExprType,
        name: Optional[Token],
        body: CodeBlock,
        """
        if node.name:
            self.emit_ln(
                node, f"except {node.ex_type.meta['py_code']} as {node.name.value}:"
            )
        else:
            self.emit_ln(node, f"except {node.ex_type.meta['py_code']}:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1

    def exit_except_list(self, node: ast.ExceptList) -> None:
        """Sub objects.

        excepts: list[Except],
        """
        for i in node.excepts:
            self.emit_ln(node, i.meta["py_code"])

    def exit_finally_stmt(self, node: ast.FinallyStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        """
        self.emit_ln(node, "finally:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1

    def exit_iter_for_stmt(self, node: ast.IterForStmt) -> None:
        """Sub objects.

        iter: Assignment,
        condition: ExprType,
        count_by: ExprType,
        body: CodeBlock,
        """
        self.emit_ln(node, f"{node.iter.meta['py_code']}")
        self.emit_ln(node, f"while {node.condition.meta['py_code']}:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.emit_ln(node, f"{node.count_by.meta['py_code']}")
        self.indent_level -= 1

    def exit_in_for_stmt(self, node: ast.InForStmt) -> None:
        """Sub objects.

        name_list: Token,
        collection: ExprType,
        body: CodeBlock,
        """
        names = node.name_list.meta["py_code"]
        self.emit_ln(node, f"for {names} in {node.collection.meta['py_code']}:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1

    def exit_name_list(self, node: ast.NameList) -> None:
        """Sub objects.

        names: list[Name],
        """
        self.emit(node, ",".join([i.meta["py_code"] for i in node.names]))

    def exit_while_stmt(self, node: ast.WhileStmt) -> None:
        """Sub objects.

        condition: ExprType,
        body: CodeBlock,
        """
        self.emit_ln(node, f"while {node.condition.meta['py_code']}:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1

    def exit_with_stmt(self, node: ast.WithStmt) -> None:
        """Sub objects.

        exprs: "ExprAsItemList",
        body: "CodeBlock",
        """
        self.emit_ln(node, f"with {node.exprs.meta['py_code']}:")
        self.indent_level += 1
        self.emit_ln(node, node.body.meta["py_code"])
        self.indent_level -= 1

    def exit_expr_as_item_list(self, node: ast.ExprAsItemList) -> None:
        """Sub objects.

        items: list["ExprAsItem"],
        """
        self.emit(node, ", ".join([i.meta["py_code"] for i in node.items]))

    def exit_expr_as_item(self, node: ast.ExprAsItem) -> None:
        """Sub objects.

        expr: "ExprType",
        alias: Optional[Token],
        """
        if node.alias:
            self.emit(node, node.expr.meta["py_code"] + " as " + node.alias.value)
        else:
            self.emit(node, node.expr.meta["py_code"])

    def exit_raise_stmt(self, node: ast.RaiseStmt) -> None:
        """Sub objects.

        cause: Optional[ExprType],
        """
        if node.cause:
            self.emit_ln(node, f"raise {node.cause.meta['py_code']}")
        else:
            self.emit_ln(node, "raise")

    def exit_assert_stmt(self, node: ast.AssertStmt) -> None:
        """Sub objects.

        condition: ExprType,
        error_msg: Optional[ExprType],
        """
        if node.error_msg:
            self.emit_ln(
                node,
                f"assert {node.condition.meta['py_code']}, {node.error_msg.meta['py_code']}",
            )
        else:
            self.emit_ln(node, f"assert {node.condition.meta['py_code']}")

    # NOTE: Incomplete for Jac Purple and Red
    def exit_ctrl_stmt(self, node: ast.CtrlStmt) -> None:
        """Sub objects.

        ctrl: Token,
        """
        if node.ctrl.name == Tok.KW_SKIP:
            self.ds_feature_warn()
        else:
            self.emit_ln(node, node.ctrl.value)

    def exit_delete_stmt(self, node: ast.DeleteStmt) -> None:
        """Sub objects.

        target: ExprType,
        """
        self.emit_ln(node, f"del {node.target.meta['py_code']}")

    # NOTE: Incomplete for Jac Purple and Red
    def exit_report_stmt(self, node: ast.ReportStmt) -> None:
        """Sub objects.

        expr: ExprType,
        """
        self.ds_feature_warn()

    def exit_return_stmt(self, node: ast.ReturnStmt) -> None:
        """Sub objects.

        expr: Optional[ExprType],
        """
        if node.expr:
            self.emit_ln(node, f"return {node.expr.meta['py_code']}")
        else:
            self.emit_ln(node, "return")

    def exit_yield_stmt(self, node: ast.YieldStmt) -> None:
        """Sub objects.

        expr: Optional[ExprType],
        """
        if node.expr:
            self.emit_ln(node, f"yield {node.expr.meta['py_code']}")
        else:
            self.emit_ln(node, "yield")

    # NOTE: Incomplete for Jac Purple and Red
    def exit_ignore_stmt(self, node: ast.IgnoreStmt) -> None:
        """Sub objects.

        target: ExprType,
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_visit_stmt(self, node: ast.VisitStmt) -> None:
        """Sub objects.

        vis_type: Optional[Token],
        target: Optional["ExprType"],
        else_body: Optional["ElseStmt"],
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_revisit_stmt(self, node: ast.RevisitStmt) -> None:
        """Sub objects.

        hops: Optional[ExprType],
        else_body: Optional[ElseStmt],
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_disengage_stmt(self, node: ast.DisengageStmt) -> None:
        """Sub objects."""
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_await_stmt(self, node: ast.AwaitStmt) -> None:
        """Sub objects.

        target: ExprType,
        """
        self.ds_feature_warn()

    def exit_assignment(self, node: ast.Assignment) -> None:
        """Sub objects.

        is_static: bool,
        target: AtomType,
        value: ExprType,
        """
        if node.is_static:
            self.warning("Static variable semantics is not supported in bootstrap Jac")
        self.emit(node, f"{node.target.meta['py_code']} = {node.value.meta['py_code']}")

    # NOTE: Incomplete for Jac Purple and Red
    def exit_binary_expr(self, node: ast.BinaryExpr) -> None:
        """Sub objects.

        left: ExprType,
        right: ExprType,
        op: Token | DisconnectOp | ConnectOp,
        """
        if isinstance(node.op, (ast.DisconnectOp, ast.ConnectOp)):
            self.ds_feature_warn()
        if isinstance(node.op, ast.Token):
            if node.op.value in [
                *["+", "-", "*", "/", "%", "**"],
                *["+=", "-=", "*=", "/=", "%=", "**="],
                *[">>", "<<", ">>=", "<<="],
                *["//=", "&=", "|=", "^=", "~="],
                *["//", "&", "|", "^"],
                *[">", "<", ">=", "<=", "==", "!=", ":="],
                *["and", "or", "in", "not in", "is", "is not"],
            ]:
                self.emit(
                    node,
                    f"{node.left.meta['py_code']} {node.op.value} {node.right.meta['py_code']}",
                )
            elif node.op.name in [
                Tok.PIPE_FWD,
                Tok.KW_SPAWN,
                Tok.A_PIPE_FWD,
            ] and isinstance(node.left, ast.TupleVal):
                params = node.left.meta["py_code"]
                params = params.replace(",)", ")") if params[-2:] == ",)" else params
                self.emit(node, f"{node.right.meta['py_code']}{params}")
            elif node.op.name in [Tok.PIPE_BKWD, Tok.A_PIPE_BKWD] and isinstance(
                node.right, ast.TupleVal
            ):
                params = node.right.meta["py_code"]
                params = params.replace(",)", ")") if params[-2:] == ",)" else params
                self.emit(node, f"{node.left.meta['py_code']}{params}")
            elif node.op.name == Tok.PIPE_FWD and isinstance(node.right, ast.TupleVal):
                self.ds_feature_warn()
            elif node.op.name == Tok.PIPE_FWD:
                self.emit(
                    node, f"{node.right.meta['py_code']}({node.left.meta['py_code']}"
                )
                paren_count = (
                    node.meta["pipe_chain_count"]
                    if "pipe_chain_count" in node.meta
                    else 1
                )
                if (
                    isinstance(node.parent, ast.BinaryExpr)
                    and isinstance(node.parent.op, ast.Token)
                    and node.parent.op.name == Tok.PIPE_FWD
                ):
                    node.parent.meta["pipe_chain_count"] = paren_count + 1
                else:
                    self.emit(node, ")" * paren_count)

            elif node.op.name in [Tok.KW_SPAWN, Tok.A_PIPE_FWD]:
                self.emit(
                    node, f"{node.right.meta['py_code']}({node.left.meta['py_code']}"
                )
                paren_count = (
                    node.meta["a_pipe_chain_count"]
                    if "a_pipe_chain_count" in node.meta
                    else 1
                )
                if (
                    isinstance(node.parent, ast.BinaryExpr)
                    and isinstance(node.parent.op, ast.Token)
                    and node.parent.op.name
                    in [
                        Tok.KW_SPAWN,
                        Tok.A_PIPE_FWD,
                    ]
                ):
                    node.parent.meta["a_pipe_chain_count"] = paren_count + 1
                else:
                    self.emit(node, ")" * paren_count)

            elif node.op.name in [Tok.PIPE_BKWD, Tok.A_PIPE_BKWD]:
                self.emit(
                    node, f"{node.left.meta['py_code']}({node.right.meta['py_code']})"
                )
            elif node.op.name == Tok.ELVIS_OP:
                self.emit(
                    node,
                    f"{Con.JAC_TMP} "
                    f"if ({Con.JAC_TMP} := ({node.left.meta['py_code']})) is not None "
                    f"else {node.right.meta['py_code']}",
                )
            else:
                self.error(
                    f"Binary operator {node.op.value} not supported in bootstrap Jac"
                )

    def exit_if_else_expr(self, node: ast.IfElseExpr) -> None:
        """Sub objects.

        condition: BinaryExpr | IfElseExpr,
        value: ExprType,
        else_value: ExprType,
        """
        self.emit(
            node,
            f"{node.value.meta['py_code']} if {node.condition.meta['py_code']} "
            f"else {node.else_value.meta['py_code']}",
        )

    def exit_unary_expr(self, node: ast.UnaryExpr) -> None:
        """Sub objects.

        operand: ExprType,
        op: Token,
        """
        if node.op.value in ["-", "~", "+"]:
            self.emit(node, f"{node.op.value}{node.operand.meta['py_code']}")
        elif node.op.value == "(":  # (expression) reuses unary expr
            self.emit(node, f"({node.operand.meta['py_code']})")
        elif node.op.value == "not":
            self.emit(node, f"not {node.operand.meta['py_code']}")
        elif node.op.name in [Tok.PIPE_FWD, Tok.KW_SPAWN, Tok.A_PIPE_FWD]:
            self.emit(node, f"{node.operand.meta['py_code']}()")
        else:
            self.error(f"Unary operator {node.op.value} not supported in bootstrap Jac")

    def exit_unpack_expr(self, node: ast.UnpackExpr) -> None:
        """Sub objects.

        target: ExprType,
        is_dict: bool,
        """
        if node.is_dict:
            self.emit(node, f"**{node.target.meta['py_code']}")
        else:
            self.emit(node, f"*{node.target.meta['py_code']}")

    def exit_multi_string(self, node: ast.MultiString) -> None:
        """Sub objects.

        strings: list[Token],
        """
        for string in node.strings:
            self.emit(node, string.meta["py_code"])

    def exit_list_val(self, node: ast.ListVal) -> None:
        """Sub objects.

        values: list[ExprType],
        """
        self.emit(
            node, f"[{', '.join([value.meta['py_code'] for value in node.values])}]"
        )

    def exit_set_val(self, node: ast.ListVal) -> None:
        """Sub objects.

        values: list[ExprType],
        """
        self.emit(
            node, f"{{{', '.join([value.meta['py_code'] for value in node.values])}}}"
        )

    def exit_tuple_val(self, node: ast.TupleVal) -> None:
        """Sub objects.

        first_expr: Optional["ExprType"],
        exprs: Optional[ExprList],
        assigns: Optional[AssignmentList],
        """
        self.emit(node, "(")
        if node.first_expr:
            self.emit(node, f"{node.first_expr.meta['py_code']}")
        if not node.exprs and not node.assigns:
            self.emit(node, ",)")
        if node.exprs:
            self.emit(node, f", {node.exprs.meta['py_code']}")

        if node.assigns:
            if node.first_expr:
                self.emit(node, f", {node.assigns.meta['py_code']}")
            else:
                self.emit(node, f"{node.assigns.meta['py_code']}")
        self.emit(node, ")")

    def exit_expr_list(self, node: ast.ExprList) -> None:
        """Sub objects.

        values: list[ExprType],
        """
        self.emit(
            node, f"{', '.join([value.meta['py_code'] for value in node.values])}"
        )

    def exit_dict_val(self, node: ast.DictVal) -> None:
        """Sub objects.

        kv_pairs: list["KVPair"],
        """
        self.emit(
            node,
            f"{{{', '.join([kv_pair.meta['py_code'] for kv_pair in node.kv_pairs])}}}",
        )

    def exit_inner_compr(self, node: ast.InnerCompr) -> None:
        """Sub objects.

        out_expr: ExprType,
        name_list: NameList,
        collection: ExprType,
        conditional: Optional[ExprType],
        is_list: bool,
        is_gen: bool,
        is_set: bool,
        """
        names = node.name_list.meta["py_code"]
        partial = (
            f"{node.out_expr.meta['py_code']} for {names} "
            f"in {node.collection.meta['py_code']}"
        )
        if node.conditional:
            partial += f" if {node.conditional.meta['py_code']}"
        if node.is_list:
            self.emit(node, f"[{partial}]")
        elif node.is_set:
            self.emit(node, f"{{{partial}}}")
        elif node.is_gen:
            self.emit(node, f"({partial})")

    def exit_dict_compr(self, node: ast.DictCompr) -> None:
        """Sub objects.

        outk_expr: ExprType,
        outv_expr: ExprType,
        name_list: NameList,
        collection: ExprType,
        conditional: Optional[ExprType],
        """
        names = node.name_list.meta["py_code"]
        partial = (
            f"{node.outk_expr.meta['py_code']}: {node.outv_expr.meta['py_code']} for "
            f"{names}"
        )
        partial += f" in {node.collection.meta['py_code']}"
        if node.conditional:
            partial += f" if {node.conditional.meta['py_code']}"
        self.emit(node, f"{{{partial}}}")

    def exit_k_v_pair(self, node: ast.KVPair) -> None:
        """Sub objects.

        key: ExprType,
        value: ExprType,
        """
        self.emit(node, f"{node.key.meta['py_code']}: {node.value.meta['py_code']}")

    def exit_atom_trailer(self, node: ast.AtomTrailer) -> None:
        """Sub objects.

        target: AtomType,
        right: IndexSlice | ArchRefType | Token,
        null_ok: bool,
        """
        if node.null_ok:
            if isinstance(node.right, ast.IndexSlice):
                self.emit(
                    node,
                    f"({node.target.meta['py_code']}{node.right.meta['py_code']} "
                    f"if {node.target.meta['py_code']} is not None else None)",
                )
            else:
                self.emit(
                    node,
                    f"({node.target.meta['py_code']}.{node.right.meta['py_code']} "
                    f"if {node.target.meta['py_code']} is not None else None)",
                )
        else:
            if isinstance(node.right, ast.IndexSlice):
                self.emit(
                    node,
                    f"{node.target.meta['py_code']}{node.right.meta['py_code']}",
                )
            else:
                self.emit(
                    node,
                    f"{node.target.meta['py_code']}.{node.right.meta['py_code']}",
                )

    # NOTE: Incomplete for Jac Purple and Red
    def exit_func_call(self, node: ast.FuncCall) -> None:
        """Sub objects.

        target: AtomType,
        params: Optional[ParamList],
        """
        if node.params:
            self.emit(
                node,
                f"{node.target.meta['py_code']}({node.params.meta['py_code']})",
            )
        else:
            self.emit(node, f"{node.target.meta['py_code']}()")

    def exit_param_list(self, node: ast.ParamList) -> None:
        """Sub objects.

        p_args: Optional[ExprList],
        p_kwargs: Optional[AssignmentList],
        """
        if node.p_args and node.p_kwargs:
            self.emit(
                node,
                f"{node.p_args.meta['py_code']}, {node.p_kwargs.meta['py_code']}",
            )
        elif node.p_args:
            self.emit(node, f"{node.p_args.meta['py_code']}")
        elif node.p_kwargs:
            self.emit(node, f"{node.p_kwargs.meta['py_code']}")

    def exit_assignment_list(self, node: ast.AssignmentList) -> None:
        """Sub objects.

        values: list[Assignment],
        """
        self.emit(
            node, f"{', '.join([value.meta['py_code'] for value in node.values])}"
        )

    def exit_index_slice(self, node: ast.IndexSlice) -> None:
        """Sub objects.

        start: ExprType,
        stop: Optional[ExprType],
        """
        if node.is_range:
            self.emit(
                node,
                f"[{node.start.meta['py_code'] if node.start else ''}:"
                f"{node.stop.meta['py_code'] if node.stop else ''}]",
            )
        elif node.start:
            self.emit(node, f"[{node.start.meta['py_code']}]")
        else:
            self.ice("Something went horribly wrong.")

    def exit_special_var_ref(self, node: ast.SpecialVarRef) -> None:
        """Sub objects.

        var: Token,
        """
        self.emit(node, node.py_resolve_name())

    # NOTE: Incomplete for Jac Purple and Red (maybe for global)
    def exit_arch_ref(self, node: ast.ArchRef) -> None:
        """Sub objects.

        name: Name,
        arch: Token,
        """
        self.emit(node, node.py_resolve_name())

    # NOTE: Incomplete for Jac Purple and Red
    def exit_edge_op_ref(self, node: ast.EdgeOpRef) -> None:
        """Sub objects.

        filter_cond: Optional[ExprType],
        edge_dir: EdgeDir,
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_disconnect_op(self, node: ast.DisconnectOp) -> None:
        """Sub objects.

        filter_cond: Optional[ExprType],
        edge_dir: EdgeDir,
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red
    def exit_connect_op(self, node: ast.ConnectOp) -> None:
        """Sub objects.

        spawn: Optional[ExprType],
        edge_dir: EdgeDir,
        """
        self.ds_feature_warn()

    # NOTE: Incomplete for Jac Purple and Red (to consider)
    def exit_filter_compr(self, node: ast.FilterCompr) -> None:
        """Sub objects.

        compares: list[BinaryExpr],
        """
        self.ds_feature_warn()

    def exit_f_string(self, node: ast.FString) -> None:
        """Sub objects.

        parts: list["Token | ExprType"],
        """
        self.emit(node, 'f"')
        for part in node.parts:
            if isinstance(part, ast.Token) and part.name == "PIECE":
                self.emit(node, f"{part.meta['py_code']}")
            else:
                self.emit(node, "{" + part.meta["py_code"] + "}")
        self.emit(node, '"')



"""Test Fstring parser."""
from typing import Callable

from jaclang.utils.fstring_parser import FStringLexer, FStringParser
from jaclang.utils.test import TestCase

fstrings = [
    ["The sum of 7 and 3 is {7 + 3}.", "7 + 3"],
    ["The lowercase version of 'HELLO' is {'HELLO'.lower()}.", "'HELLO'.lower()"],
    [
        "My name is {my_dict['name']} and I am {my_dict['age']} years old.",
        "my_dict['name']",
        "my_dict['age']",
    ],
    ["My favorite fruit is {my_list[0]}.", "my_list[0]"],
    [
        "The sum of 5 and 7 is {(lambda x, y: x + y)(5, 7)}.",
        "(lambda x, y: x + y)(5, 7)",
    ],
    ["The person's name is {person.name}.", "person.name"],
    [
        "{'left aligned':<15} | {'center':^10} | {'right aligned':>15}",
        "'left aligned':<15",
        "'center':^10",
        "'right aligned':>15",
    ],
    ["Pi to 2 decimal places is {3.14159:.2f}.", "3.14159:.2f"],
    ["Integer with leading zeros {42:05}.", "42:05"],
    ["Current date is {datetime.now():%Y-%m-%d}.", "datetime.now():%Y-%m-%d"],
]


class TestFstrings(TestCase):
    """Test fstring parser."""

    pass  # Remove existing test case


def make_test_func(example: list[str]) -> Callable:
    """Generate a new test function for each example in fstrings."""

    def test_func(self) -> None:  # noqa: ANN001
        test_string = 'f"' + example[0] + '"'
        tree = FStringParser().parse(FStringLexer().tokenize(test_string))
        for i in example[1:]:
            self.assertTrue(f"'{i}'" in str(tree) or f'"{i}"' in str(tree))
        self.assertEqual(str(tree).count("EXPR_START"), len(example) - 1)

    return test_func


# Generate a new test case for each example in fstrings
for idx, example in enumerate(fstrings):
    # Attach the new test function to the test class
    setattr(TestFstrings, f"test_fstring_lexer_{idx}", make_test_func(example))



"""Utility functions and classes for Jac compilation toolchain."""
import re

import jaclang.jac.absyntree as ast


def pascal_to_snake(pascal_string: str) -> str:
    """Convert pascal case to snake case."""
    snake_string = re.sub(r"(?<!^)(?=[A-Z])", "_", pascal_string).lower()
    return snake_string


def add_line_numbers(s: str) -> str:
    """Add line numbers to a string."""
    lines = s.split("\n")
    return "\n".join(f"{i+1}: \t{line}" for i, line in enumerate(lines))


def clip_code_section(s: str, target_line: int, line_range: int) -> str:
    """Clip a section of code and highlight target line."""
    lines = s.split("\n")
    start = max(0, target_line - line_range - 1)
    end = min(target_line + line_range, len(lines))

    result = []
    for i in range(start, end):
        line = lines[i]
        if i == target_line - 1:
            line = "*" + line
        result.append(line)
    return "\n".join(result)


def get_ast_nodes_as_snake_case() -> list[str]:
    """Get all AST nodes as snake case."""
    import inspect
    import sys

    module_name = ast.__name__
    module = sys.modules[module_name]

    # Retrieve the source code of the module
    source_code = inspect.getsource(module)

    classes = inspect.getmembers(module, inspect.isclass)
    ast_node_classes = [cls for _, cls in classes if issubclass(cls, ast.AstNode)]

    ordered_classes = sorted(
        ast_node_classes, key=lambda cls: source_code.find(f"class {cls.__name__}")
    )
    snake_names = []
    for cls in ordered_classes:
        class_name = cls.__name__
        snake_names.append(pascal_to_snake(class_name))
    return snake_names



"""Tests for Jac parser."""
from jaclang.jac.absyntree import AstNode
from jaclang.jac.lexer import JacLexer
from jaclang.jac.parser import JacParser
from jaclang.utils.test import TestCaseMicroSuite


class TestParser(TestCaseMicroSuite):
    """Test Jac self.prse."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        lex = JacLexer(mod_path=filename, input_ir=self.file_to_str(filename)).ir
        prse = JacParser(mod_path=filename, input_ir=lex)
        self.assertFalse(prse.errors_had)

    def test_shift_reduce_conflict(self) -> None:
        """Test for shift reduce conflict."""
        self.assertEqual(len(JacParser._lrtable.sr_conflicts), 0)

    def test_reduce_reduce_conflict(self) -> None:
        """Test for reduce reduce conflict."""
        self.assertEqual(len(JacParser._lrtable.rr_conflicts), 0)

    def test_basci_parsing(self) -> None:
        """Basic test for parsing."""
        lex = JacLexer(mod_path="fam.jac", input_ir=self.load_fixture("fam.jac")).ir
        prse = JacParser(mod_path="fam.jac", input_ir=lex)
        output = prse.ir
        self.assertFalse(prse.errors_had)
        if isinstance(output, AstNode):
            self.assertGreater(len(str(output.to_dict())), 1000)
        else:
            self.fail("Output is not an AstNode.")

    def test_parsing_keyword_escape(self) -> None:
        """Basic test for parsing."""
        lex = JacLexer(mod_path="kwesc.jac", input_ir=self.load_fixture("kwesc.jac")).ir
        prse = JacParser(mod_path="kwesc.jac", input_ir=lex)
        output = prse.ir
        self.assertFalse(prse.errors_had)
        self.assertIn("'name': 'KWESC_NAME', 'value': 'walker'", str(output.to_dict()))

    def test_parsing_jac_cli(self) -> None:
        """Basic test for parsing."""
        lex = JacLexer(
            mod_path="../../../cli/cli.jac",
            input_ir=self.load_fixture("../../../cli/cli.jac"),
        ).ir
        prse = JacParser(mod_path="../../../cli/cli.jac", input_ir=lex)
        self.assertFalse(prse.errors_had)


TestParser.self_attach_micro_tests()



"""Type Analyze Pass."""
from types import ModuleType
from typing import Any

import jaclang.jac.absyntree as ast
from jaclang.jac.passes import Pass
from jaclang.jac.sym_table import SymbolTable, TypedSymbol


class TypeAnalyzePass(Pass, SymbolTable):
    """Type inference and checking pass."""

    def before_pass(self) -> None:
        """Initialize pass."""
        self.sym_tab = SymbolTable(scope_name="global")

    def exit_parse(self, node: ast.Parse) -> None:
        """Sub objects.

        name: str,
        """

    def exit_token(self, node: ast.Token) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        """

    def exit_name(self, node: ast.Name) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        already_declared: bool,
        """

    def exit_constant(self, node: ast.Constant) -> None:
        """Sub objects.

        name: str,
        value: str,
        col_start: int,
        col_end: int,
        typ: type,
        """

    def exit_module(self, node: ast.Module) -> None:
        """Sub objects.

        name: str,
        doc: Token,
        body: Elements,
        mod_path: str,
        """

    def exit_elements(self, node: ast.Elements) -> None:
        """Sub objects.

        elements: list['GlobalVars | Test | ModuleCode | Import | Architype | Ability | AbilitySpec'],
        """

    def exit_global_vars(self, node: ast.GlobalVars) -> None:
        """Sub objects.

        doc: Optional[DocString],
        access: Optional[Token],
        assignments: AssignmentList,
        """
        if node.access:
            for i in self.get_all_sub_nodes(node, typ=ast.Assignment):
                if isinstance(i.target, ast.Name) and not self.sym_tab.update(
                    TypedSymbol(
                        name=i.target.value,
                        access=node.access.value,
                        node=i,
                        typ=i._typ,
                    )
                ):
                    self.ice(
                        f"ICE: Variable {i.target.value} not seen as declared in pass {self.__class__.__name__}"
                    )

    def exit_test(self, node: ast.Test) -> None:
        """Sub objects.

        name: Name,
        doc: Optional[DocString],
        description: Token,
        body: CodeBlock,
        """

    def exit_module_code(self, node: ast.ModuleCode) -> None:
        """Sub objects.

        doc: Optional[DocString],
        body: CodeBlock,
        """

    def exit_import(self, node: ast.Import) -> None:
        """Sub objects.

        lang: Name,
        path: ModulePath,
        alias: Optional[Name],
        items: Optional[ModuleItems],
        is_absorb: bool,
        self.sub_module = None
        """
        if node.lang.value == "jac" and not node.sub_module:
            self.ice(f"Jac module not loaded by pass {self.__class__.__name__}")
        if node.items:
            for i in node.items.items:
                if i.alias:
                    self.sym_tab.set(
                        TypedSymbol(
                            name=i.alias.value,
                            typ=Any,  # TODO: Backpatch analysis for module itmes
                            node=node,
                        ),
                        fresh_only=True,
                    )
                else:
                    self.sym_tab.set(
                        TypedSymbol(
                            name=i.name.value,
                            typ=Any,  # TODO: Backpatch analysis for module itmes
                            node=node,
                        ),
                        fresh_only=True,
                    )
        else:
            self.sym_tab.set(
                TypedSymbol(
                    name=node.path.path[-1].value,
                    typ=ModuleType,
                    node=node,
                ),
                fresh_only=True,
            )

    def exit_module_path(self, node: ast.ModulePath) -> None:
        """Sub objects.

        path: list[Token],
        """

    def exit_module_item(self, node: ast.ModuleItem) -> None:
        """Sub objects.

        name: Name,
        alias: Optional[Token],
        self.body: Optional[AstNode] = None
        """

    def exit_module_items(self, node: ast.ModuleItems) -> None:
        """Sub objects.

        items: list['ModuleItem'],
        """

    def exit_architype(self, node: ast.Architype) -> None:
        """Sub objects.

        name: Name,
        arch_type: Token,
        doc: Optional[DocString],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[ArchBlock],
        """
        node._typ = type
        exists = self.sym_tab.lookup(name=node.name.value, deep=False)
        # TODO: if exists and type(exists.def_node) == ast.ArchDef and not node.body: # This should be own pass
        #     node.body = exists.def_node.body
        if exists and not self.assert_type_match(sym=exists, node=node):
            return
        self.sym_tab.set(
            TypedSymbol(
                name=node.name.value,
                typ=type,
                node=node,
                access=node.access.value if node.access else None,
            ),
        )

    def exit_arch_def(self, node: ast.ArchDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        mod: Optional[DottedNameList],
        arch: ObjectRef | NodeRef | EdgeRef | WalkerRef,
        body: ArchBlock,
        """

    def exit_decorators(self, node: ast.Decorators) -> None:
        """Sub objects.

        calls: list['ExprType'],
        """

    def exit_base_classes(self, node: ast.BaseClasses) -> None:
        """Sub objects.

        base_classes: list['DottedNameList'],
        """

    def exit_ability(self, node: ast.Ability) -> None:
        """Sub objects.

        name: Name,
        is_func: bool,
        doc: Optional[DocString],
        decorators: Optional[Decorators],
        access: Optional[Token],
        signature: FuncSignature | TypeSpec | EventSignature,
        body: CodeBlock,
        """

    def exit_ability_def(self, node: ast.AbilityDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        mod: Optional[DottedNameList],
        ability: AbilityRef,
        body: CodeBlock,
        """

    def exit_arch_block(self, node: ast.ArchBlock) -> None:
        """Sub objects.

        members: list['ArchHas | Ability'],
        """

    def exit_arch_has(self, node: ast.ArchHas) -> None:
        """Sub objects.

        doc: Optional[DocString],
        access: Optional[Token],
        vars: HasVarList,
        """

    def exit_has_var(self, node: ast.HasVar) -> None:
        """Sub objects.

        name: Name,
        type_tag: TypeSpec,
        value: Optional[ExprType],
        """

    def exit_has_var_list(self, node: ast.HasVarList) -> None:
        """Sub objects.

        vars: list['HasVar'],
        """

    def exit_type_spec_list(self, node: ast.TypeSpecList) -> None:
        """Sub objects.

        types: list[TypeSpec],
        """

    def exit_type_spec(self, node: ast.TypeSpec) -> None:
        """Sub objects.

        typ: Token | DottedNameList,
        list_nest: TypeSpec,
        dict_nest: TypeSpec,
        """

    def exit_event_signature(self, node: ast.EventSignature) -> None:
        """Sub objects.

        event: Token,
        arch_tag_info: Optional[DottedNameList | Token],
        """

    def exit_dotted_name_list(self, node: ast.DottedNameList) -> None:
        """Sub objects.

        names: list[Token],
        dotted: bool,
        """

    def exit_func_signature(self, node: ast.FuncSignature) -> None:
        """Sub objects.

        params: Optional[FuncParams],
        return_type: Optional[TypeSpec],
        """

    def exit_func_params(self, node: ast.FuncParams) -> None:
        """Sub objects.

        params: list['ParamVar'],
        """

    def exit_param_var(self, node: ast.ParamVar) -> None:
        """Sub objects.

        name: Name,
        unpack: Optional[Token],
        type_tag: TypeSpec,
        value: Optional[ExprType],
        """

    def exit_enum(self, node: ast.Enum) -> None:
        """Sub objects.

        name: Name,
        doc: Optional[DocString],
        decorators: Optional[Decorators],
        access: Optional[Token],
        base_classes: BaseClasses,
        body: Optional[EnumBlock],
        """

    def exit_enum_def(self, node: ast.EnumDef) -> None:
        """Sub objects.

        doc: Optional[DocString],
        mod: Optional[DottedNameList],
        body: EnumBlock,
        """

    def exit_enum_block(self, node: ast.EnumBlock) -> None:
        """Sub objects.

        stmts: list['Name|Assignment'],
        """

    def exit_code_block(self, node: ast.CodeBlock) -> None:
        """Sub objects.

        stmts: list['StmtType'],
        """

    def exit_typed_ctx_block(self, node: ast.TypedCtxBlock) -> None:
        """Sub objects.

        type_ctx: TypeList,
        body: CodeBlock,
        """

    def exit_if_stmt(self, node: ast.IfStmt) -> None:
        """Sub objects.

        condition: ExprType,
        body: CodeBlock,
        elseifs: Optional[ElseIfs],
        else_body: Optional[ElseStmt],
        """

    def exit_else_ifs(self, node: ast.ElseIfs) -> None:
        """Sub objects.

        elseifs: list['IfStmt'],
        """

    def exit_else_stmt(self, node: ast.ElseStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        """

    def exit_try_stmt(self, node: ast.TryStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        excepts: Optional[ExceptList],
        finally_body: Optional[FinallyStmt],
        """

    def exit_except(self, node: ast.Except) -> None:
        """Sub objects.

        typ: ExprType,
        name: Optional[Token],
        body: CodeBlock,
        """

    def exit_except_list(self, node: ast.ExceptList) -> None:
        """Sub objects.

        excepts: list['Except'],
        """

    def exit_finally_stmt(self, node: ast.FinallyStmt) -> None:
        """Sub objects.

        body: CodeBlock,
        """

    def exit_iter_for_stmt(self, node: ast.IterForStmt) -> None:
        """Sub objects.

        iter: Assignment,
        condition: ExprType,
        count_by: ExprType,
        body: CodeBlock,
        """

    def exit_in_for_stmt(self, node: ast.InForStmt) -> None:
        """Sub objects.

        name: Name,
        collection: ExprType,
        body: CodeBlock,
        """

    def exit_name_list(self, node: ast.NameList) -> None:
        """Sub objects.

        names: list[Name],
        """

    def exit_while_stmt(self, node: ast.WhileStmt) -> None:
        """Sub objects.

        condition: ExprType,
        body: CodeBlock,
        """

    def exit_with_stmt(self, node: ast.WithStmt) -> None:
        """Sub objects.

        exprs: "ExprAsItemList",
        body: "CodeBlock",
        """

    def exit_expr_as_item_list(self, node: ast.ExprAsItemList) -> None:
        """Sub objects.

        items: list["ExprAsItem"],
        """

    def exit_expr_as_item(self, node: ast.ExprAsItem) -> None:
        """Sub objects.

        expr: "ExprType",
        alias: Optional[Token],
        """

    def exit_raise_stmt(self, node: ast.RaiseStmt) -> None:
        """Sub objects.

        cause: Optional[ExprType],
        """

    def exit_assert_stmt(self, node: ast.AssertStmt) -> None:
        """Sub objects.

        condition: ExprType,
        error_msg: Optional[ExprType],
        """

    def exit_ctrl_stmt(self, node: ast.CtrlStmt) -> None:
        """Sub objects.

        ctrl: Token,
        """

    def exit_delete_stmt(self, node: ast.DeleteStmt) -> None:
        """Sub objects.

        target: ExprType,
        """

    def exit_report_stmt(self, node: ast.ReportStmt) -> None:
        """Sub objects.

        expr: ExprType,
        """

    def exit_return_stmt(self, node: ast.ReturnStmt) -> None:
        """Sub objects.

        expr: Optional[ExprType],
        """

    def exit_yield_stmt(self, node: ast.YieldStmt) -> None:
        """Sub objects.

        expr: Optional[ExprType],
        """

    def exit_ignore_stmt(self, node: ast.IgnoreStmt) -> None:
        """Sub objects.

        target: ExprType,
        """

    def exit_visit_stmt(self, node: ast.VisitStmt) -> None:
        """Sub objects.

        typ: Optional[Token],
        target: Optional[ExprType],
        else_body: Optional[ElseStmt],
        """

    def exit_revisit_stmt(self, node: ast.RevisitStmt) -> None:
        """Sub objects.

        hops: Optional[ExprType],
        else_body: Optional[ElseStmt],
        """

    def exit_disengage_stmt(self, node: ast.DisengageStmt) -> None:
        """Sub objects."""

    def exit_await_stmt(self, node: ast.AwaitStmt) -> None:
        """Sub objects.

        target: ExprType,
        """

    def exit_assignment(self, node: ast.Assignment) -> None:
        """Sub objects.

        is_static: bool,
        target: AtomType,
        value: ExprType,
        """

    def exit_binary_expr(self, node: ast.BinaryExpr) -> None:
        """Sub objects.

        left: ExprType,
        right: ExprType,
        op: Token,
        """

    def exit_if_else_expr(self, node: ast.IfElseExpr) -> None:
        """Sub objects.

        condition: BinaryExpr | IfElseExpr,
        value: ExprType,
        else_value: ExprType,
        """

    def exit_unary_expr(self, node: ast.UnaryExpr) -> None:
        """Sub objects.

        operand: ExprType,
        op: Token,
        """

    def exit_unpack_expr(self, node: ast.UnpackExpr) -> None:
        """Sub objects.

        target: ExprType,
        is_dict: bool,
        """

    def exit_multi_string(self, node: ast.MultiString) -> None:
        """Sub objects.

        strings: list['Token | FString'],
        """

    def exit_list_val(self, node: ast.ListVal) -> None:
        """Sub objects.

        values: list['ExprType'],
        """

    def exit_set_val(self, node: ast.ListVal) -> None:
        """Sub objects.

        values: list[ExprType],
        """

    def exit_tuple_val(self, node: ast.TupleVal) -> None:
        """Sub objects.

        first_expr: Optional["ExprType"],
        exprs: Optional[ExprList],
        assigns: Optional[AssignmentList],
        """

    def exit_expr_list(self, node: ast.ExprList) -> None:
        """Sub objects.

        values: list['ExprType'],
        """

    def exit_dict_val(self, node: ast.DictVal) -> None:
        """Sub objects.

        kv_pairs: list['KVPair'],
        """

    def exit_inner_compr(self, node: ast.InnerCompr) -> None:
        """Sub objects.

        out_expr: "ExprType",
        name: Name,
        collection: "ExprType",
        conditional: Optional["ExprType"],
        is_list: bool,
        is_gen: bool,
        is_set: bool,
        """

    def exit_dict_compr(self, node: ast.DictCompr) -> None:
        """Sub objects.

        outk_expr: ExprType,
        outv_expr: ExprType,
        k_name: Name,
        v_name: Optional[Token],
        collection: ExprType,
        conditional: Optional[ExprType],
        """

    def exit_k_v_pair(self, node: ast.KVPair) -> None:
        """Sub objects.

        key: ExprType,
        value: ExprType,
        """

    def exit_atom_trailer(self, node: ast.AtomTrailer) -> None:
        """Sub objects.

        target: AtomType,
        right: IndexSlice | ArchRefType | Token,
        null_ok: bool,
        """

    def exit_func_call(self, node: ast.FuncCall) -> None:
        """Sub objects.

        target: AtomType,
        params: Optional[ParamList],
        """

    def exit_param_list(self, node: ast.ParamList) -> None:
        """Sub objects.

        p_args: Optional[ExprList],
        p_kwargs: Optional[AssignmentList],
        """

    def exit_assignment_list(self, node: ast.AssignmentList) -> None:
        """Sub objects.

        values: list['ExprType'],
        """

    def exit_index_slice(self, node: ast.IndexSlice) -> None:
        """Sub objects.

        start: ExprType,
        stop: Optional[ExprType],
        """

    def exit_arch_ref(self, node: ast.ArchRef) -> None:
        """Sub objects.

        name: Name,
        """

    def exit_special_var_ref(self, node: ast.SpecialVarRef) -> None:
        """Sub objects.

        name: Optional[Token],
        """

    def exit_edge_op_ref(self, node: ast.EdgeOpRef) -> None:
        """Sub objects.

        filter_cond: Optional[ExprType],
        edge_dir: EdgeDir,
        """

    def exit_disconnect_op(self, node: ast.DisconnectOp) -> None:
        """Sub objects.

        filter_cond: Optional[ExprType],
        edge_dir: EdgeDir,
        """

    def exit_connect_op(self, node: ast.ConnectOp) -> None:
        """Sub objects.

        spawn: Optional[ExprType],
        edge_dir: EdgeDir,
        """

    def exit_filter_compr(self, node: ast.FilterCompr) -> None:
        """Sub objects.

        compares: list[BinaryExpr],
        """

    def exit_f_string(self, node: ast.FString) -> None:
        """Sub objects.

        parts: list['Token | ExprType'],
        """

    # Checks and validations
    # ----------------------

    def assert_type_match(self, sym: TypedSymbol, node: ast.AstNode) -> bool:
        """Check if two types match."""
        if isinstance(sym.typ, node._typ):
            return True
        self.error(
            f"Type mismatch, {sym.name} already defined on line "
            f"{sym.node.line if sym.node else 'unknown'} as {sym.typ} not compatible with"
            f"{node._typ} on line {node.line}!"
        )
        return False

    def not_defined_err(self, name: str) -> None:
        """Check if a symbol is defined."""
        self.error(f"Symbol {name} not defined!")

    def already_defined_err(self, name: str) -> None:
        """Check if a symbol is defined."""
        self.error(f"Symbol {name} already in use in this scope!")



"""Lexer for Jac language."""
from typing import Generator

from jaclang.jac.transform import ABCLexerMeta, Transform
from jaclang.utils.sly.lex import Lexer, Token


class JacLexer(Lexer, Transform, metaclass=ABCLexerMeta):
    """Jac Lexer."""

    def __init__(
        self,
        mod_path: str,
        input_ir: str,
        base_path: str = "",
        prior: Transform | None = None,
    ) -> None:
        """Initialize lexer."""
        Transform.__init__(self, mod_path, input_ir, base_path, prior)  # type: ignore
        self.ir: Generator = self.ir

    tokens = {
        "FLOAT",
        "STRING",
        "DOC_STRING",
        "FSTRING",
        "BOOL",
        "INT",
        "HEX",
        "BIN",
        "OCT",
        "NULL",
        "NAME",
        "KWESC_NAME",
        "TYP_STRING",
        "TYP_INT",
        "TYP_FLOAT",
        "TYP_LIST",
        "TYP_TUPLE",
        "TYP_SET",
        "TYP_DICT",
        "TYP_BOOL",
        "TYP_BYTES",
        "TYP_ANY",
        "TYP_TYPE",
        "KW_FREEZE",
        "KW_OBJECT",
        "KW_ENUM",
        "KW_NODE",
        "KW_IGNORE",
        "KW_VISIT",
        "KW_REVISIT",
        "KW_SPAWN",
        "KW_WITH",
        "KW_ENTRY",
        "KW_EXIT",
        "KW_IMPORT",
        "KW_INCLUDE",
        "KW_FROM",
        "KW_AS",
        "KW_EDGE",
        "KW_WALKER",
        "KW_ASYNC",
        "KW_AWAIT",
        "KW_TEST",
        "KW_ASSERT",
        "COLON",
        "PIPE_FWD",
        "PIPE_BKWD",
        "DOT_FWD",
        "DOT_BKWD",
        "LBRACE",
        "RBRACE",
        "SEMI",
        "EQ",
        "ADD_EQ",
        "SUB_EQ",
        "MUL_EQ",
        "FLOOR_DIV_EQ",
        "DIV_EQ",
        "MOD_EQ",
        "BW_AND_EQ",
        "BW_OR_EQ",
        "BW_XOR_EQ",
        "BW_NOT_EQ",
        "LSHIFT_EQ",
        "RSHIFT_EQ",
        "WALRUS_EQ",
        "KW_AND",
        "KW_OR",
        "KW_IF",
        "KW_ELIF",
        "KW_ELSE",
        "KW_FOR",
        "KW_TO",
        "KW_BY",
        "KW_WHILE",
        "KW_CONTINUE",
        "KW_BREAK",
        "KW_DISENGAGE",
        "KW_YIELD",
        "KW_SKIP",
        "KW_REPORT",
        "KW_RETURN",
        "KW_DELETE",
        "KW_TRY",
        "KW_EXCEPT",
        "KW_FINALLY",
        "KW_RAISE",
        "DOT",
        "NOT",
        "EE",
        "LT",
        "GT",
        "LTE",
        "GTE",
        "NE",
        "KW_IN",
        "KW_IS",
        "KW_NIN",
        "KW_ISN",
        "KW_PRIV",
        "KW_PUB",
        "KW_PROT",
        "KW_HAS",
        "KW_GLOBAL",
        "COMMA",
        "KW_CAN",
        "KW_STATIC",
        "PLUS",
        "MINUS",
        "STAR_MUL",
        "FLOOR_DIV",
        "DIV",
        "MOD",
        "BW_AND",
        "BW_OR",
        "BW_XOR",
        "BW_NOT",
        "LSHIFT",
        "RSHIFT",
        "STAR_POW",
        "LPAREN",
        "RPAREN",
        "LSQUARE",
        "RSQUARE",
        "ARROW_L",
        "ARROW_R",
        "ARROW_BI",
        "ARROW_L_p1",
        "ARROW_L_p2",
        "ARROW_R_p1",
        "ARROW_R_p2",
        "CARROW_L",
        "CARROW_R",
        "CARROW_L_p1",
        "CARROW_L_p2",
        "CARROW_R_p1",
        "CARROW_R_p2",
        "GLOBAL_OP",
        "HERE_OP",
        "SELF_OP",
        "INIT_OP",
        "SUPER_OP",
        "ROOT_OP",
        "WALKER_OP",
        "NODE_OP",
        "EDGE_OP",
        "OBJECT_OP",
        "ENUM_OP",
        "ABILITY_OP",
        "A_PIPE_FWD",
        "A_PIPE_BKWD",
        "ELVIS_OP",
        "RETURN_HINT",
        "NULL_OK",
        "DECOR_OP",
    }

    # Ignored patterns
    ignore_ws = r"[ \t]+"
    ignore_newline = r"[\r\n]+"  # type: ignore
    ignore_comment = r"#\*(.|\n|\r)*?\*#"  # type: ignore
    ignore_py_comment = r"#.*"

    # Regular expression rules for tokens
    FLOAT = r"(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?"
    DOC_STRING = r'"""(.|\n|\r)*?"""|\'\'\'(.|\n|\r)*?\'\'\''  # type: ignore
    FSTRING = r'f"[^"\r\n]*"|f\'[^\'\r\n]*\''
    STRING = r'"[^"\r\n]*"|\'[^\'\r\n]*\''
    BOOL = r"True|False"
    KW_NIN = r"\bnot\s+in\b"
    KW_ISN = r"\bis\s+not\b"
    HEX = r"0[xX][0-9a-fA-F_]+"
    BIN = r"0[bB][01_]+"
    OCT = r"0[oO][0-7_]+"
    INT = r"[0-9][0-9_]*"
    NULL = r"None"
    KWESC_NAME = r"<>[a-zA-Z_][a-zA-Z0-9_]*"
    NAME = r"[a-zA-Z_][a-zA-Z0-9_]*"

    # Keywords
    NAME["str"] = "TYP_STRING"  # type: ignore
    NAME["int"] = "TYP_INT"  # type: ignore
    NAME["float"] = "TYP_FLOAT"  # type: ignore
    NAME["list"] = "TYP_LIST"  # type: ignore
    NAME["tuple"] = "TYP_TUPLE"  # type: ignore
    NAME["set"] = "TYP_SET"  # type: ignore
    NAME["dict"] = "TYP_DICT"  # type: ignore
    NAME["bool"] = "TYP_BOOL"  # type: ignore
    NAME["bytes"] = "TYP_BYTES"  # type: ignore
    NAME["any"] = "TYP_ANY"  # type: ignore
    NAME["type"] = "TYP_TYPE"  # type: ignore
    NAME["freeze"] = "KW_FREEZE"  # type: ignore
    NAME["object"] = "KW_OBJECT"  # type: ignore
    NAME["enum"] = "KW_ENUM"  # type: ignore
    NAME["node"] = "KW_NODE"  # type: ignore
    NAME["ignore"] = "KW_IGNORE"  # type: ignore
    NAME["visit"] = "KW_VISIT"  # type: ignore
    NAME["revisit"] = "KW_REVISIT"  # type: ignore
    NAME["spawn"] = "KW_SPAWN"  # type: ignore
    NAME["with"] = "KW_WITH"  # type: ignore
    NAME["entry"] = "KW_ENTRY"  # type: ignore
    NAME["exit"] = "KW_EXIT"  # type: ignore
    NAME["import"] = "KW_IMPORT"  # type: ignore
    NAME["include"] = "KW_INCLUDE"  # type: ignore
    NAME["from"] = "KW_FROM"  # type: ignore
    NAME["as"] = "KW_AS"  # type: ignore
    NAME["edge"] = "KW_EDGE"  # type: ignore
    NAME["walker"] = "KW_WALKER"  # type: ignore
    NAME["async"] = "KW_ASYNC"  # type: ignore
    NAME["await"] = "KW_AWAIT"  # type: ignore
    NAME["test"] = "KW_TEST"  # type: ignore
    NAME["assert"] = "KW_ASSERT"  # type: ignore
    NAME["and"] = "KW_AND"  # type: ignore
    NAME["or"] = "KW_OR"  # type: ignore
    NAME["if"] = "KW_IF"  # type: ignore
    NAME["elif"] = "KW_ELIF"  # type: ignore
    NAME["else"] = "KW_ELSE"  # type: ignore
    NAME["for"] = "KW_FOR"  # type: ignore
    NAME["to"] = "KW_TO"  # type: ignore
    NAME["by"] = "KW_BY"  # type: ignore
    NAME["while"] = "KW_WHILE"  # type: ignore
    NAME["continue"] = "KW_CONTINUE"  # type: ignore
    NAME["break"] = "KW_BREAK"  # type: ignore
    NAME["disengage"] = "KW_DISENGAGE"  # type: ignore
    NAME["yield"] = "KW_YIELD"  # type: ignore
    NAME["skip"] = "KW_SKIP"  # type: ignore
    NAME["report"] = "KW_REPORT"  # type: ignore
    NAME["return"] = "KW_RETURN"  # type: ignore
    NAME["del"] = "KW_DELETE"  # type: ignore
    NAME["try"] = "KW_TRY"  # type: ignore
    NAME["except"] = "KW_EXCEPT"  # type: ignore
    NAME["finally"] = "KW_FINALLY"  # type: ignore
    NAME["raise"] = "KW_RAISE"  # type: ignore
    NAME["in"] = "KW_IN"  # type: ignore
    NAME["is"] = "KW_IS"  # type: ignore
    NAME["not"] = "NOT"  # type: ignore
    NAME["private"] = "KW_PRIV"  # type: ignore
    NAME["public"] = "KW_PUB"  # type: ignore
    NAME["protected"] = "KW_PROT"  # type: ignore
    NAME["has"] = "KW_HAS"  # type: ignore
    NAME["global"] = "KW_GLOBAL"  # type: ignore
    NAME["can"] = "KW_CAN"  # type: ignore
    NAME["static"] = "KW_STATIC"  # type: ignore

    # Special Arrow Tokens
    ARROW_L = r"<--"
    ARROW_R = r"-->"
    ARROW_BI = r"<-->"
    ARROW_L_p1 = r"<-\["
    ARROW_L_p2 = r"]-"
    ARROW_R_p1 = r"-\["
    ARROW_R_p2 = r"]->"
    CARROW_L = r"<\+\+"
    CARROW_R = r"\+\+>"
    CARROW_L_p1 = r"<\+\["
    CARROW_R_p2 = r"]\+>"
    CARROW_L_p2 = r"]\+"
    CARROW_R_p1 = r"\+\["

    # Just special
    GLOBAL_OP = r":g:|:global:"
    HERE_OP = r"<h>|<here>"
    SELF_OP = r"<s>|<self>"
    INIT_OP = r"<i>|<init>"
    SUPER_OP = r"<super>"
    ROOT_OP = r"<r>|<root>"
    WALKER_OP = r":w:|:walker:"
    NODE_OP = r":n:|:node:"
    EDGE_OP = r":e:|:edge:"
    OBJECT_OP = r":o:|:object:"
    ENUM_OP = r":enum:"
    ABILITY_OP = r":a:|:ability:"
    A_PIPE_FWD = r":>"
    A_PIPE_BKWD = r"<:"
    PIPE_FWD = r"\|>"
    PIPE_BKWD = r"<\|"
    DOT_FWD = r"\.>"
    DOT_BKWD = r"<\."
    RETURN_HINT = r"->"
    ELVIS_OP = r"\?:"
    NULL_OK = r"\?"
    DECOR_OP = r"@"

    # Token rules
    KW_AND = r"&&"
    KW_OR = r"\|\|"
    ADD_EQ = r"\+="
    SUB_EQ = r"-="
    MUL_EQ = r"\*="
    FLOOR_DIV_EQ = r"//="
    DIV_EQ = r"/="
    MOD_EQ = r"%="
    BW_AND_EQ = r"&="
    BW_OR_EQ = r"\|="
    BW_XOR_EQ = r"\^="
    BW_NOT_EQ = r"~="
    LSHIFT_EQ = r"<<="
    RSHIFT_EQ = r">>="
    LSHIFT = r"<<"
    RSHIFT = r">>"
    LTE = r"<="
    GTE = r">="
    NE = r"!="
    NOT = r"!"
    WALRUS_EQ = r":="
    COLON = r":"
    LBRACE = r"{"
    RBRACE = r"}"
    SEMI = r";"
    EE = r"=="
    EQ = r"="
    DOT = r"\."
    LT = r"<"
    GT = r">"
    COMMA = r","
    PLUS = r"\+"
    MINUS = r"-"
    STAR_POW = r"\*\*"
    STAR_MUL = r"\*"
    FLOOR_DIV = r"//"
    DIV = r"/"
    MOD = r"%"
    BW_AND = r"&"
    BW_OR = r"\|"
    BW_XOR = r"\^"
    BW_NOT = r"~"
    LPAREN = r"\("
    RPAREN = r"\)"
    LSQUARE = r"\["
    RSQUARE = r"\]"

    def ignore_newline(self, t: Token) -> Token:  # noqa
        """Increment line number."""
        self.lineno += len(t.value)
        return t

    def ignore_comment(self, t: Token) -> Token:  # noqa
        """Add docstring to lexer."""
        self.lineno += t.value.count("\n")
        self.lineno += t.value.count("\r")
        return t

    def DOC_STRING(self, t: Token) -> Token:  # noqa
        """Add docstring to lexer."""
        self.lineno += t.value.count("\n")
        self.lineno += t.value.count("\r")
        return t

    # Transform Implementations
    # -------------------------
    def transform(self, ir: str) -> Generator:
        """Tokenize the input."""
        return self.tokenize(ir)

    def error(self, t: Token) -> None:
        """Raise an error for illegal characters."""
        self.cur_line = self.lineno
        self.log_error(msg=f"Illegal character '{t.value[0]}'")
        self.index += 1



"""Core primitives for Jaseci."""
from jaclang import jac_blue_import

prim = jac_blue_import("primitives")


Object = prim.Object
Node = prim.Node
Edge = prim.Edge
Walker = prim.Walker
make_architype = prim.make_architype

exec_ctx = prim.exec_ctx



"""Test pass module."""
from jaclang.jac.passes.blue import DeclDefMatchPass
from jaclang.jac.transpiler import jac_file_to_pass
from jaclang.utils.test import TestCase


class DeclDefMatchPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_import_values_avail(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("base.jac"), "", DeclDefMatchPass
        )
        self.assertFalse(state.errors_had)
        self.assertIn("mine", state.sym_tab.tab)
        self.assertIsNotNone(state.sym_tab.tab["mine"].node.body)

    def test_ability_connected_to_decl(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("base.jac"), "", DeclDefMatchPass
        )
        self.assertFalse(state.errors_had)
        self.assertIn("Test.say_hi", state.sym_tab.tab)
        self.assertIsNotNone(state.sym_tab.tab["Test.say_hi"].node.body)
        self.assertIn("Test.init", state.sym_tab.tab)
        self.assertIsNotNone(state.sym_tab.tab["Test.init"].node.body)

    def test_ability_connected_to_decl_post(self) -> None:
        """Basic test for pass."""
        state = jac_file_to_pass(
            self.fixture_abs_path("base2.jac"), "", DeclDefMatchPass
        )
        self.assertFalse(state.errors_had)
        self.assertIn("Test.say_hi", state.sym_tab.tab)
        self.assertIsNotNone(state.sym_tab.tab["Test.say_hi"].node.body)
        self.assertIn("Test.init", state.sym_tab.tab)
        self.assertIsNotNone(state.sym_tab.tab["Test.init"].node.body)

    def test_collision_error_correct(self) -> None:
        """Basic test for multi defs."""
        state = jac_file_to_pass(
            self.fixture_abs_path("decls.jac"), "", DeclDefMatchPass
        )
        self.assertTrue(state.errors_had)
        self.assertIn("/impl/defs2.jac", state.errors_had[0])
        self.assertIn("/impl/defs1.jac", state.errors_had[0])
        self.assertIn("/impl/defs2.jac", state.errors_had[1])
        self.assertIn("/impl/defs1.jac", state.errors_had[1])



"""Language tools for the Jaclang project."""

import inspect
import sys
from typing import Optional

import jaclang.jac.absyntree as ast
from jaclang.jac.parser import JacLexer
from jaclang.utils.helpers import pascal_to_snake


class AstKidInfo:
    """Information about a kid."""

    def __init__(self, name: str, typ: str, default: Optional[str] = None) -> None:
        """Initialize."""
        self.name = name
        self.typ = typ
        self.default = default


class AstNodeInfo:
    """Meta data about AST nodes."""

    type_map: dict[str, type] = {}

    def __init__(self, cls: type) -> None:
        """Initialize."""
        self.cls = cls
        self.process(cls)

    def process(self, cls: type) -> None:
        """Process AstNode class."""
        self.name = cls.__name__
        self.doc = cls.__doc__
        AstNodeInfo.type_map[self.name] = cls
        self.class_name_snake = pascal_to_snake(cls.__name__)
        self.init_sig = inspect.signature(cls.__init__)
        self.kids: list[AstKidInfo] = []
        for param_name, param in self.init_sig.parameters.items():
            if param_name not in ["self", "parent", "kid", "line", "mod_link"]:
                param_type = (
                    param.annotation
                    if param.annotation != inspect.Parameter.empty
                    else "Any"
                )
                param_default = (
                    param.default if param.default != inspect.Parameter.empty else None
                )
                self.kids.append(AstKidInfo(param_name, param_type, param_default))


class AstTool:
    """Ast tools."""

    def __init__(self) -> None:
        """Initialize."""
        module = sys.modules[ast.__name__]
        source_code = inspect.getsource(module)
        classes = inspect.getmembers(module, inspect.isclass)
        ast_node_classes = [
            AstNodeInfo(cls) for _, cls in classes if issubclass(cls, ast.AstNode)
        ]
        self.ast_classes = sorted(
            ast_node_classes,
            key=lambda cls: source_code.find(f"class {cls.name}"),
        )

    def pass_template(self) -> str:
        """Generate pass template."""
        output = "import jaclang.jac.absyntree as ast\nfrom jaclang.jac.passes import Pass\n\nclass SomePass(Pass):\n"

        def emit(to_append: str) -> None:
            """Emit to output."""
            nonlocal output
            output += "\n    " + to_append

        for cls in self.ast_classes:
            emit(
                f"def exit_{cls.class_name_snake}(self, node: ast.{cls.name}) -> None:\n"
            )
            emit('    """Sub objects.\n')

            for kid in cls.kids:
                emit(
                    f"    {kid.name}: {kid.typ}{' ='+kid.default if kid.default else ''},"
                )

            emit('    """\n')
        output = (
            output.replace("jaclang.jac.absyntree.", "")
            .replace("typing.", "")
            .replace("<enum '", "")
            .replace("'>", "")
            .replace("<class '", "")
            .replace("ForwardRef('", "")
            .replace("')", "")
        )
        return output

    def jac_keywords(self) -> str:
        """Get all Jac keywords as an or string."""
        ret = ""
        for k in JacLexer._remapping["NAME"].keys():
            ret += f"{k}|"
        return ret[:-1]

    def md_doc(self) -> str:
        """Generate mermaid markdown doc."""
        output = ""
        for cls in self.ast_classes:
            if not len(cls.kids):
                continue
            output += f"## {cls.name}\n"
            output += "```mermaid\nflowchart LR\n"
            for kid in cls.kids:
                if "_end" in kid.name:
                    kid.name = kid.name.replace("_end", "_end_")
                arrow = "-.->" if "Optional" in kid.typ else "-->"
                typ = (
                    kid.typ.replace("Optional[", "")
                    .replace("]", "")
                    .replace("|", ",")
                    .replace("list[", "list - ")
                )
                output += f"{cls.name} {arrow}|{typ}| {kid.name}\n"
            output += "```\n\n"
            output += f"{cls.doc} \n\n"
        return output



"""Collection of purple passes for Jac IR."""
from jaclang import jac_blue_import
from jaclang.jac.passes.blue import pass_schedule


purple = jac_blue_import("purple_pygen_pass")
analyze = jac_blue_import("analyze_pass")
AnalyzePass = analyze.AnalyzePass  # type: ignore
PurplePygenPass = purple.PurplePygenPass  # type: ignore

__all__ = [
    "AnalyzePass",
    "PurplePygenPass",
]


pass_schedule = [
    *pass_schedule[:-1],
    AnalyzePass,
    PurplePygenPass,
]



"""Test type analyze pass module."""
import inspect

from jaclang.jac.passes.blue import BluePygenPass, TypeAnalyzePass
from jaclang.jac.transpiler import jac_file_to_pass
from jaclang.utils.helpers import get_ast_nodes_as_snake_case as ast_snakes
from jaclang.utils.test import TestCaseMicroSuite


class TypeAnalyzePassTests(TestCaseMicroSuite):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pygen_jac_cli(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("../../../../../cli/cli.jac"), target=BluePygenPass
        )
        # print(code_gen.ir.meta["py_code"])
        self.assertFalse(code_gen.errors_had)
        self.assertGreater(len(code_gen.ir.meta["py_code"]), 200)

    def test_pass_ast_complete(self) -> None:
        """Test for enter/exit name diffs with parser."""
        ast_func_names = [
            x for x in ast_snakes() if x not in ["ast_node", "o_o_p_access_node"]
        ]
        pygen_func_names = []
        for name, value in inspect.getmembers(TypeAnalyzePass):
            if (
                (name.startswith("enter_") or name.startswith("exit_"))
                and inspect.isfunction(value)
                and not getattr(TypeAnalyzePass.__base__, value.__name__, False)
                and value.__qualname__.split(".")[0]
                == TypeAnalyzePass.__name__.replace("enter_", "").replace("exit_", "")
            ):
                pygen_func_names.append(name.replace("enter_", "").replace("exit_", ""))
        for name in pygen_func_names:
            self.assertIn(name, ast_func_names)
        for name in ast_func_names:
            self.assertIn(name, pygen_func_names)

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        ast = jac_file_to_pass(filename, "", target=BluePygenPass).ir
        typed_ast = TypeAnalyzePass(mod_path=filename, input_ir=ast).ir
        self.assertGreater(len(str(typed_ast.to_dict())), 10)


TypeAnalyzePassTests.self_attach_micro_tests()



"""Test ast build pass module."""
import inspect

from jaclang.jac.lexer import JacLexer
from jaclang.jac.parser import JacParser
from jaclang.jac.passes.blue import AstBuildPass
from jaclang.utils.fstring_parser import FStringParser
from jaclang.utils.test import TestCaseMicroSuite


class AstBuildPassTests(TestCaseMicroSuite):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pass_grammar_complete(self) -> None:
        """Test for enter/exit name diffs with parser."""
        from jaclang.jac.parser import JacParser

        parser_func_names = []
        for name, value in [
            *inspect.getmembers(JacParser),
            *inspect.getmembers(FStringParser),
        ]:
            if (
                inspect.isfunction(value)
                and value.__qualname__.split(".")[0]
                in [JacParser.__name__, FStringParser.__name__]
                and name not in ["__init__", "error", "transform"]
            ):
                parser_func_names.append(name)
        ast_build_func_names = []
        for name, value in inspect.getmembers(AstBuildPass):
            if (
                (name.startswith("enter_") or name.startswith("exit_"))
                and inspect.isfunction(value)
                and not getattr(AstBuildPass.__base__, value.__name__, False)
                and value.__qualname__.split(".")[0]
                == AstBuildPass.__name__.replace("enter_", "").replace("exit_", "")
            ):
                ast_build_func_names.append(
                    name.replace("enter_", "").replace("exit_", "")
                )
        for name in ast_build_func_names:
            self.assertIn(name, parser_func_names)
        for name in parser_func_names:
            self.assertIn(name, ast_build_func_names)

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        lex = JacLexer(mod_path=f"{filename}", input_ir=self.file_to_str(filename)).ir
        prse = JacParser(mod_path=f"{filename}", input_ir=lex).ir
        build_pass = AstBuildPass(mod_path="", input_ir=prse).ir
        self.assertIsNotNone(build_pass)
        if build_pass:
            self.assertGreater(len(str(build_pass.to_dict())), 200)


AstBuildPassTests.self_attach_micro_tests()



"""Transpilation functions."""
from typing import Type, TypeVar

import jaclang.jac.absyntree as ast
from jaclang.jac.parser import JacLexer
from jaclang.jac.parser import JacParser
from jaclang.jac.passes import Pass
from jaclang.jac.passes.blue import BluePygenPass, pass_schedule
from jaclang.jac.transform import Transform


T = TypeVar("T", bound=Pass)


def jac_file_to_parse_tree(file_path: str, base_dir: str) -> Transform:
    """Convert a Jac file to an AST."""
    with open(file_path) as file:
        lex = JacLexer(mod_path=file_path, input_ir=file.read(), base_path=base_dir)
        prse = JacParser(
            mod_path=file_path, input_ir=lex.ir, base_path=base_dir, prior=lex
        )
        return prse


def transpile_jac_blue(file_path: str, base_dir: str) -> str:
    """Transpiler Jac file and return python code as string."""
    code = jac_file_to_pass(
        file_path=file_path, base_dir=base_dir, target=BluePygenPass
    )
    if isinstance(code.ir, ast.Module):
        return code.ir.meta["py_code"]
    else:
        raise code.gen_exception("Transpilation of Jac file failed.")


def transpile_jac_purple(file_path: str, base_dir: str) -> str:
    """Transpiler Jac file and return python code as string."""
    from jaclang.jac.passes.purple import pass_schedule, PurplePygenPass

    code = jac_file_to_pass(
        file_path=file_path,
        base_dir=base_dir,
        target=PurplePygenPass,
        schedule=pass_schedule,
    )
    if isinstance(code.ir, ast.Module):
        return code.ir.meta["py_code"]
    else:
        raise code.gen_exception("Transpilation of Jac file failed.")


def jac_file_to_pass(
    file_path: str,
    base_dir: str = "",
    target: Type[T] = BluePygenPass,
    schedule: list[Type[T]] = pass_schedule,
) -> T:
    """Convert a Jac file to an AST."""
    ast_ret = jac_file_to_parse_tree(file_path, base_dir)
    for i in schedule:
        if i == target:
            break
        ast_ret = i(
            mod_path=file_path, input_ir=ast_ret.ir, base_path=base_dir, prior=ast_ret
        )
    ast_ret = target(
        mod_path=file_path, input_ir=ast_ret.ir, base_path=base_dir, prior=ast_ret
    )
    return ast_ret



"""Special Imports for Jac Code."""
import inspect
import marshal
import sys
import traceback
import types
from os import makedirs, path
from typing import Callable, Optional

from jaclang.jac.constant import Constants as Con, Values as Val
from jaclang.jac.transpiler import transpile_jac_blue, transpile_jac_purple
from jaclang.utils.helpers import add_line_numbers, clip_code_section


def import_jac_module(
    transpiler_func: Callable,
    target: str,
    base_path: Optional[str] = None,
    cachable: bool = True,
) -> Optional[types.ModuleType]:
    """Core Import Process."""
    target = path.join(*(target.split("."))) + ".jac"

    dir_path, file_name = path.split(target)
    module_name = path.splitext(file_name)[0]
    package_path = dir_path.replace(path.sep, ".")

    if package_path and f"{package_path}.{module_name}" in sys.modules:
        return sys.modules[f"{package_path}.{module_name}"]
    elif not package_path and module_name in sys.modules:
        return sys.modules[module_name]

    if base_path:
        caller_dir = path.dirname(base_path) if not path.isdir(base_path) else base_path
    else:
        frame = inspect.stack()[2]
        caller_dir = path.dirname(path.abspath(frame[0].f_code.co_filename))
    full_target = path.normpath(path.join(caller_dir, target))

    dev_dir = path.join(caller_dir, "__jac_gen__")
    makedirs(dev_dir, exist_ok=True)
    py_file_path = path.join(dev_dir, module_name + ".py")
    if (
        cachable
        and path.exists(py_file_path)
        and path.getmtime(py_file_path) > path.getmtime(full_target)
    ):
        with open(py_file_path, "r") as f:
            code_string = f.read()
    else:
        code_string = transpiler_func(file_path=full_target, base_dir=caller_dir)
        with open(py_file_path, "w") as f:
            f.write(code_string)

    module = types.ModuleType(module_name)
    module.__file__ = full_target
    module.__name__ = module_name
    module.__dict__["_jac_pycodestring_"] = code_string

    if (
        cachable
        and path.exists(py_file_path + "c")
        and path.getmtime(py_file_path + "c") > path.getmtime(full_target)
    ):
        with open(py_file_path + "c", "rb") as f:
            codeobj = marshal.load(f)
    else:
        codeobj = compile(code_string, f"_jac_py_gen ({module.__file__})", "exec")
        with open(py_file_path + "c", "wb") as f:
            marshal.dump(codeobj, f)

    try:
        exec(codeobj, module.__dict__)
    except Exception as e:
        tb = traceback.extract_tb(e.__traceback__)
        err = handle_jac_error(code_string, e, tb)
        raise type(e)(str(e) + "\n" + err)

    if package_path:
        parts = package_path.split(".")
        for i in range(len(parts)):
            package_name = ".".join(parts[: i + 1])
            if package_name not in sys.modules:
                sys.modules[package_name] = types.ModuleType(package_name)

        setattr(sys.modules[package_path], module_name, module)
        sys.modules[f"{package_path}.{module_name}"] = module
    else:
        sys.modules[module_name] = module
    return module


def handle_jac_error(code_string: str, e: Exception, tb: traceback.StackSummary) -> str:
    """Handle Jac Error."""
    except_line = e.end_lineno if isinstance(e, SyntaxError) else list(tb)[-1].lineno
    if not isinstance(except_line, int) or except_line == 0:
        return ""
    py_error_region = clip_code_section(
        add_line_numbers(code_string), except_line, Val.JAC_ERROR_LINE_RANGE
    )
    try:
        jac_err_line = int(code_string.splitlines()[except_line - 1].split()[-1])
        mod_index = int(code_string.splitlines()[except_line - 1].split()[-2])
        mod_paths = code_string.split(Con.JAC_DEBUG_SPLITTER)[1].strip().splitlines()
        with open(mod_paths[mod_index], "r") as file:
            jac_code_string = file.read()
        jac_error_region = clip_code_section(
            add_line_numbers(jac_code_string), jac_err_line, Val.JAC_ERROR_LINE_RANGE
        )
    except Exception as e:
        jac_error_region = str(e)
    snippet = (
        f"JacCode Snippet:\n{jac_error_region}\n"
        f"PyCode Snippet:\n{py_error_region}\n"
    )
    return snippet


def jac_blue_import(
    target: str, base_path: Optional[str] = None, cachable: bool = True
) -> Optional[types.ModuleType]:
    """Jac Blue Imports."""
    return import_jac_module(transpile_jac_blue, target, base_path, cachable)


def jac_purple_import(
    target: str, base_path: Optional[str] = None, cachable: bool = True
) -> Optional[types.ModuleType]:
    """Jac Purple Imports."""
    return import_jac_module(transpile_jac_purple, target, base_path, cachable)



"""Standardized transformation process and error interface."""
from __future__ import annotations

import os
from abc import ABC, ABCMeta, abstractmethod
from typing import Optional


from jaclang.jac.absyntree import AstNode
from jaclang.utils.log import logging
from jaclang.utils.sly.lex import LexerMeta
from jaclang.utils.sly.yacc import ParserMeta


class TransformError(Exception):
    """Error during transformation."""

    def __init__(self, message: str, errors: list[str], warnings: list[str]) -> None:
        """Initialize error."""
        self.errors = errors
        self.warnings = warnings
        if len(errors):
            message += "\nErrors:"
            for i in self.errors:
                message += "\n" + i
        if len(warnings):
            message += "\nWarnings:"
            for i in self.warnings:
                message += "\n" + i
        super().__init__(message)


class Transform(ABC):
    """Abstract class for IR passes."""

    def __init__(
        self,
        mod_path: str,
        input_ir: AstNode,
        base_path: str = "",
        prior: Optional[Transform] = None,
    ) -> None:
        """Initialize pass."""
        self.logger = logging.getLogger(self.__class__.__module__)
        self.errors_had = [] if not prior else prior.errors_had
        self.warnings_had = [] if not prior else prior.warnings_had
        self.cur_line = 0
        self.mod_path = mod_path
        self.rel_mod_path = (
            mod_path.replace(base_path, "") if base_path else mod_path.split(os.sep)[-1]
        )
        self.ir = self.transform(ir=input_ir)

    @abstractmethod
    def transform(self, ir: AstNode) -> AstNode:
        """Transform interface."""
        pass

    def log_error(self, msg: str) -> None:
        """Pass Error."""
        msg = f"Mod {self.rel_mod_path}: Line {self.cur_line}, " + msg
        self.errors_had.append(msg)
        self.logger.error(msg)

    def log_warning(self, msg: str) -> None:
        """Pass Error."""
        msg = f"Mod {self.rel_mod_path}: Line {self.cur_line}, " + msg
        self.warnings_had.append(msg)
        self.logger.warning(msg)

    def gen_exception(
        self, msg: str = "Error in code transform, see above for details."
    ) -> TransformError:
        """Raise error."""
        return TransformError(msg, self.errors_had, self.warnings_had)


class ABCLexerMeta(ABCMeta, LexerMeta):
    """Metaclass for Jac Lexer."""

    pass


class ABCParserMeta(ABCMeta, ParserMeta):
    """Metaclass for Jac Lexer."""

    pass



"""Test ast build pass module."""
import inspect

from jaclang.jac.passes.blue import BluePygenPass
from jaclang.jac.transpiler import jac_file_to_pass, transpile_jac_blue
from jaclang.utils.helpers import get_ast_nodes_as_snake_case as ast_snakes
from jaclang.utils.test import TestCaseMicroSuite


class BluePygenPassTests(TestCaseMicroSuite):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_jac_cli(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("../../../../../cli/cli.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)

    def test_pipe_operator(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn(
            'say((dump(print(len)))({"name": "value"}))', code_gen.ir.meta["py_code"]
        )
        self.assertIn(
            '{"name": "value"}(len(print(print(print))))', code_gen.ir.meta["py_code"]
        )
        self.assertIn("a = (5 + 10) * 2", code_gen.ir.meta["py_code"])

    def test_atomic_pipe_operator(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn(
            'say((dump(print)(len))({"name": "value"}))', code_gen.ir.meta["py_code"]
        )

    def test_pipe_operator_multi_param(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn("self.func(*args, **kwargs)", code_gen.ir.meta["py_code"])
        self.assertIn("inspect.signature(func)", code_gen.ir.meta["py_code"])
        self.assertIn("self.registry.items()", code_gen.ir.meta["py_code"])

    def test_with_stmt(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn(
            'with open("file.txt") as f, open("file2.txt") as f:',
            code_gen.ir.meta["py_code"],
        )

    def test_empty_codeblock(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn("pass", code_gen.ir.meta["py_code"])

    def test_enum_gen(self) -> None:
        """Basic test for pass."""
        code_gen = jac_file_to_pass(
            self.fixture_abs_path("codegentext.jac"), target=BluePygenPass
        )
        self.assertFalse(code_gen.errors_had)
        self.assertIn(
            "from enum import Enum as __jac_Enum__, auto as __jac_auto__",
            code_gen.ir.meta["py_code"],
        )
        self.assertIn("class Color(__jac_Enum__):", code_gen.ir.meta["py_code"])
        self.assertIn("GREEN = __jac_auto__()", code_gen.ir.meta["py_code"])
        self.assertIn("RED = 1", code_gen.ir.meta["py_code"])

    def test_pass_ast_complete(self) -> None:
        """Test for enter/exit name diffs with parser."""
        ast_func_names = [
            x for x in ast_snakes() if x not in ["ast_node", "o_o_p_access_node"]
        ]
        pygen_func_names = []
        for name, value in inspect.getmembers(BluePygenPass):
            if (
                (name.startswith("enter_") or name.startswith("exit_"))
                and inspect.isfunction(value)
                and not getattr(BluePygenPass.__base__, value.__name__, False)
                and value.__qualname__.split(".")[0]
                == BluePygenPass.__name__.replace("enter_", "").replace("exit_", "")
            ):
                pygen_func_names.append(name.replace("enter_", "").replace("exit_", ""))
        for name in pygen_func_names:
            self.assertIn(name, ast_func_names)
        for name in ast_func_names:
            self.assertIn(name, pygen_func_names)

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        code_gen = transpile_jac_blue(filename, "")
        self.assertGreater(len(code_gen), 10)


BluePygenPassTests.self_attach_micro_tests()



# type: ignore
"""Parser for Jac."""
from typing import Generator, Optional

import jaclang.jac.absyntree as ast
from jaclang.jac.lexer import JacLexer
from jaclang.jac.transform import ABCParserMeta, Transform
from jaclang.utils.sly.yacc import Parser, YaccProduction


_ = None  # For flake8 linting


class JacParser(Transform, Parser, metaclass=ABCParserMeta):
    """Parser for Jac."""

    start = "module"

    def __init__(
        self,
        mod_path: str,
        input_ir: Generator,
        base_path: str = "",
        prior: Optional[Transform] = None,
    ) -> None:
        """Initialize parser."""
        Transform.__init__(self, mod_path, input_ir, base_path, prior)
        self.ir_tup = self.ir
        if not self.errors_had:
            self.ir: ast.AstNode = parse_tree_to_ast(self.ir)

    tokens = JacLexer.tokens
    # debugfile = "parser.out"

    # All mighty start rule
    # ---------------------
    @_(
        "DOC_STRING",
        "DOC_STRING element_list",
    )
    def module(self, p: YaccProduction) -> YaccProduction:
        """Start rule."""
        return p

    # Jac program structured as a list of elements
    # --------------------------------------------
    @_(
        "element",
        "element_list element",
    )
    def element_list(self, p: YaccProduction) -> YaccProduction:
        """Element list rule."""
        return p

    # Element types
    # -------------
    @_(
        "global_var",
        "test",
        "mod_code",
        "import_stmt",
        "include_stmt",
        "architype",
        "ability",
        "enum",
    )
    def element(self, p: YaccProduction) -> YaccProduction:
        """Element rule."""
        return p

    @_(
        "doc_tag KW_GLOBAL access_tag assignment_list SEMI",
        "doc_tag KW_FREEZE access_tag assignment_list SEMI",
    )
    def global_var(self, p: YaccProduction) -> YaccProduction:
        """Global variable rule."""
        return p

    @_(
        "KW_PRIV",
        "KW_PUB",
        "KW_PROT",
    )
    def access(self, p: YaccProduction) -> YaccProduction:
        """Permission tag rule."""
        return p

    @_(
        "COLON access",
        "empty",
    )
    def access_tag(self, p: YaccProduction) -> YaccProduction:
        """Permission tag rule."""
        return p

    @_("doc_tag KW_TEST NAME multistring code_block")
    def test(self, p: YaccProduction) -> YaccProduction:
        """Test rule."""
        return p

    @_("doc_tag KW_WITH KW_ENTRY code_block")
    def mod_code(self, p: YaccProduction) -> YaccProduction:
        """Module-level free code rule."""
        return p

    @_(
        "empty",
        "DOC_STRING",
    )
    def doc_tag(self, p: YaccProduction) -> YaccProduction:
        """Doc tag rule."""
        return p

    # Import Statements
    # -----------------
    @_(
        "KW_IMPORT sub_name import_path SEMI",
        "KW_IMPORT sub_name import_path KW_AS NAME SEMI",
        "KW_IMPORT sub_name KW_FROM import_path COMMA import_items SEMI",
    )
    def import_stmt(self, p: YaccProduction) -> YaccProduction:
        """Import rule."""
        return p

    @_("KW_INCLUDE sub_name import_path SEMI")
    def include_stmt(self, p: YaccProduction) -> YaccProduction:
        """Import rule."""
        return p

    @_(
        "import_path_prefix",
        "import_path_prefix import_path_tail",
    )
    def import_path(self, p: YaccProduction) -> YaccProduction:
        """Import path rule."""
        return p

    @_(
        "esc_name",
        "DOT esc_name",
        "DOT DOT esc_name",
    )
    def import_path_prefix(self, p: YaccProduction) -> YaccProduction:
        """Import path prefix rule."""
        return p

    @_(
        "DOT esc_name",
        "import_path_tail DOT esc_name",
    )
    def import_path_tail(self, p: YaccProduction) -> YaccProduction:
        """Import path tail rule."""
        return p

    @_(
        "named_refs",
        "named_refs KW_AS NAME",
        "import_items COMMA named_refs",
        "import_items COMMA named_refs KW_AS NAME",
    )
    def import_items(self, p: YaccProduction) -> YaccProduction:
        """Name as list rule."""
        return p

    # Architype elements
    # ------------------
    @_(
        "architype_decl",
        "architype_def",
    )
    def architype(self, p: YaccProduction) -> YaccProduction:
        """Architype rule."""
        return p

    @_(
        "doc_tag arch_type access_tag NAME inherited_archs SEMI",
        "doc_tag decorators arch_type access_tag NAME inherited_archs SEMI",
        "doc_tag arch_type access_tag NAME inherited_archs member_block",
        "doc_tag decorators arch_type access_tag NAME inherited_archs member_block",
    )
    def architype_decl(self, p: YaccProduction) -> YaccProduction:
        """Architype declaration rule."""
        return p

    @_(
        "doc_tag strict_arch_ref member_block",
        "doc_tag dotted_name strict_arch_ref member_block",
    )
    def architype_def(self, p: YaccProduction) -> YaccProduction:
        """Architype definition rule."""
        return p

    @_(
        "KW_NODE",
        "KW_EDGE",
        "KW_OBJECT",
        "KW_WALKER",
    )
    def arch_type(self, p: YaccProduction) -> YaccProduction:
        """Arch type rule."""
        return p

    @_(
        "DECOR_OP atom",
        "decorators DECOR_OP atom",
    )
    def decorators(self, p: YaccProduction) -> YaccProduction:
        """Python style decorator rule."""
        return p

    @_(
        "empty",
        "inherited_archs sub_name_dotted",
    )
    def inherited_archs(self, p: YaccProduction) -> YaccProduction:
        """Sub name list rule."""
        return p

    @_("COLON NAME")
    def sub_name(self, p: YaccProduction) -> YaccProduction:
        """Sub name rule."""
        return p

    @_("COLON dotted_name")
    def sub_name_dotted(self, p: YaccProduction) -> YaccProduction:
        """Sub name rule."""
        return p

    @_(
        "all_refs",
        "dotted_name DOT all_refs",
    )
    def dotted_name(self, p: YaccProduction) -> YaccProduction:
        """Strict arch reference rule."""
        return p

    @_(
        "NAME",
        "KWESC_NAME",
    )
    def esc_name(self, p: YaccProduction) -> YaccProduction:
        """Escaped name rule."""
        return p

    @_(
        "named_refs",
        "special_refs",
    )
    def all_refs(self, p: YaccProduction) -> YaccProduction:
        """All reference rules."""
        return p

    @_(
        "esc_name",
        "arch_ref",
    )
    def named_refs(self, p: YaccProduction) -> YaccProduction:
        """All reference rules."""
        return p

    @_(
        "HERE_OP",
        "SELF_OP",
        "SUPER_OP",
        "ROOT_OP",
        "INIT_OP",
    )
    def special_refs(self, p: YaccProduction) -> YaccProduction:
        """All reference rules."""
        return p

    # Ability elements
    # ----------------
    @_(
        "ability_decl",
        "KW_ASYNC ability_decl",
        "ability_def",
    )
    def ability(self, p: YaccProduction) -> YaccProduction:
        """Ability rule."""
        return p

    @_(
        "doc_tag static_tag KW_CAN access_tag all_refs event_clause SEMI",
        "doc_tag static_tag KW_CAN access_tag all_refs func_decl SEMI",
        "doc_tag static_tag KW_CAN access_tag all_refs event_clause code_block",
        "doc_tag static_tag KW_CAN access_tag all_refs func_decl code_block",
        "ability_decl_decor",
    )
    def ability_decl(self, p: YaccProduction) -> YaccProduction:
        """Ability rule."""
        return p

    @_(
        "doc_tag decorators static_tag KW_CAN access_tag all_refs event_clause SEMI",
        "doc_tag decorators static_tag KW_CAN access_tag all_refs func_decl SEMI",
        "doc_tag decorators static_tag KW_CAN access_tag all_refs event_clause code_block",
        "doc_tag decorators static_tag KW_CAN access_tag all_refs func_decl code_block",
    )
    def ability_decl_decor(self, p: YaccProduction) -> YaccProduction:
        """Ability declaration rule."""
        return p

    @_(
        "doc_tag ability_ref event_clause code_block",
        "doc_tag dotted_name ability_ref event_clause code_block",
        "doc_tag ability_ref func_decl code_block",
        "doc_tag dotted_name ability_ref func_decl code_block",
    )
    def ability_def(self, p: YaccProduction) -> YaccProduction:
        """Ability rule."""
        return p

    @_(
        "KW_WITH KW_ENTRY return_type_tag",
        "KW_WITH KW_EXIT return_type_tag",
        "KW_WITH type_spec KW_ENTRY return_type_tag",
        "KW_WITH type_spec KW_EXIT return_type_tag",
    )
    def event_clause(self, p: YaccProduction) -> YaccProduction:
        """Event clause rule."""
        return p

    @_(
        "return_type_tag",
        "LPAREN RPAREN return_type_tag",
        "LPAREN func_decl_param_list RPAREN return_type_tag",
    )
    def func_decl(self, p: YaccProduction) -> YaccProduction:
        """Func declaration parameter rule."""
        return p

    @_(
        "param_var",
        "func_decl_param_list COMMA param_var",
    )
    def func_decl_param_list(self, p: YaccProduction) -> YaccProduction:
        """Func declaration parameters list rule."""
        return p

    @_(
        "NAME type_tag",
        "NAME type_tag EQ expression",
        "STAR_MUL NAME type_tag",
        "STAR_MUL NAME type_tag EQ expression",
        "STAR_POW NAME type_tag",
        "STAR_POW NAME type_tag EQ expression",
    )
    def param_var(self, p: YaccProduction) -> YaccProduction:
        """Parameter variable rule rule."""
        return p

    # Enum elements
    # ----------------
    @_(
        "enum_decl",
        "enum_def",
    )
    def enum(self, p: YaccProduction) -> YaccProduction:
        """Enum rule."""
        return p

    @_(
        "doc_tag KW_ENUM access_tag NAME inherited_archs SEMI",
        "doc_tag KW_ENUM access_tag NAME inherited_archs enum_block",
        "doc_tag decorators KW_ENUM access_tag NAME inherited_archs SEMI",
        "doc_tag decorators KW_ENUM access_tag NAME inherited_archs enum_block",
    )
    def enum_decl(self, p: YaccProduction) -> YaccProduction:
        """Enum decl rule."""
        return p

    @_(
        "doc_tag enum_ref enum_block",
        "doc_tag dotted_name enum_ref enum_block",
    )
    def enum_def(self, p: YaccProduction) -> YaccProduction:
        """Enum def rule."""
        return p

    @_(
        "LBRACE RBRACE",
        "LBRACE enum_stmt_list RBRACE",
    )
    def enum_block(self, p: YaccProduction) -> YaccProduction:
        """Enum block rule."""
        return p

    @_(
        "NAME",
        "enum_op_assign",
        "enum_stmt_list COMMA NAME",
        "enum_stmt_list COMMA enum_op_assign",
    )
    def enum_stmt_list(self, p: YaccProduction) -> YaccProduction:
        """Enum op list rule."""
        return p

    @_(
        "NAME EQ expression",
    )
    def enum_op_assign(self, p: YaccProduction) -> YaccProduction:
        """Enum op assign rule."""
        return p

    # Attribute blocks
    # ----------------
    @_(
        "LBRACE RBRACE",
        "LBRACE member_stmt_list RBRACE",
    )
    def member_block(self, p: YaccProduction) -> YaccProduction:
        """Attribute block rule."""
        return p

    @_(
        "member_stmt",
        "member_stmt_list member_stmt",
    )
    def member_stmt_list(self, p: YaccProduction) -> YaccProduction:
        """Attribute statement list rule."""
        return p

    @_(
        "has_stmt",
        "ability",
    )
    def member_stmt(self, p: YaccProduction) -> YaccProduction:
        """Attribute statement rule."""
        return p

    # Has statements
    # --------------
    @_(
        "doc_tag static_tag KW_HAS access_tag has_assign_clause SEMI",
        "doc_tag static_tag KW_FREEZE access_tag has_assign_clause SEMI",
    )
    def has_stmt(self, p: YaccProduction) -> YaccProduction:
        """Has statement rule."""
        return p

    @_(
        "empty",
        "KW_STATIC",
    )
    def static_tag(self, p: YaccProduction) -> YaccProduction:
        """KW_Static tag rule."""
        return p

    @_(
        "typed_has_clause",
        "has_assign_clause COMMA typed_has_clause",
    )
    def has_assign_clause(self, p: YaccProduction) -> YaccProduction:
        """Has assign list rule."""
        return p

    @_(
        "esc_name type_tag",
        "esc_name type_tag EQ expression",
    )
    def typed_has_clause(self, p: YaccProduction) -> YaccProduction:
        """Parameter variable rule rule."""
        return p

    @_("COLON type_spec")
    def type_tag(self, p: YaccProduction) -> YaccProduction:
        """Type hint rule."""
        return p

    @_(
        "empty",
        "RETURN_HINT type_spec",
    )
    def return_type_tag(self, p: YaccProduction) -> YaccProduction:
        """Type hint rule."""
        return p

    @_(
        "single_type",
        "type_spec NULL_OK",
        "type_spec BW_OR single_type",
    )
    def type_spec(self, p: YaccProduction) -> YaccProduction:
        """Type hint rule."""
        return p

    @_(
        "builtin_type",
        "NULL",
        "dotted_name",
        "TYP_LIST LSQUARE single_type RSQUARE",
        "TYP_TUPLE LSQUARE single_type RSQUARE",
        "TYP_SET LSQUARE single_type RSQUARE",
        "TYP_DICT LSQUARE single_type COMMA single_type RSQUARE",
    )
    def single_type(self, p: YaccProduction) -> YaccProduction:
        """Type hint rule."""
        return p

    @_(
        "TYP_STRING",
        "TYP_BYTES",
        "TYP_INT",
        "TYP_FLOAT",
        "TYP_LIST",
        "TYP_TUPLE",
        "TYP_SET",
        "TYP_DICT",
        "TYP_BOOL",
        "TYP_ANY",
        "TYP_TYPE",
    )
    def builtin_type(self, p: YaccProduction) -> YaccProduction:
        """Any type rule."""
        return p

    # Codeblock statements
    # --------------------
    @_(
        "LBRACE RBRACE",
        "LBRACE statement_list RBRACE",
    )
    def code_block(self, p: YaccProduction) -> YaccProduction:
        """Code block rule."""
        return p

    @_(
        "statement_list statement",
        "statement",
    )
    def statement_list(self, p: YaccProduction) -> YaccProduction:
        """Statement list rule."""
        return p

    @_(
        "import_stmt",
        "architype_decl",
        "ability_decl",
        "typed_ctx_block",
        "assignment SEMI",
        "static_assignment",
        "expression SEMI",
        "if_stmt",
        "try_stmt",
        "for_stmt",
        "while_stmt",
        "with_stmt",
        "raise_stmt SEMI",
        "assert_stmt SEMI",
        "ctrl_stmt SEMI",
        "delete_stmt SEMI",
        "report_stmt SEMI",
        "return_stmt SEMI",
        "yield_stmt SEMI",
        "await_stmt SEMI",
        "walker_stmt",
    )
    def statement(self, p: YaccProduction) -> YaccProduction:
        """Statement rule."""
        return p

    @_("RETURN_HINT type_spec code_block")
    def typed_ctx_block(self, p: YaccProduction) -> YaccProduction:
        """Typed context block rule."""
        return p

    @_(
        "KW_IF expression code_block",
        "KW_IF expression code_block else_stmt",
        "KW_IF expression code_block elif_list",
        "KW_IF expression code_block elif_list else_stmt",
    )
    def if_stmt(self, p: YaccProduction) -> YaccProduction:
        """If statement rule."""
        return p

    @_(
        "KW_ELIF expression code_block",
        "elif_list KW_ELIF expression code_block",
    )
    def elif_list(self, p: YaccProduction) -> YaccProduction:
        """Else if statement list rule."""
        return p

    @_("KW_ELSE code_block")
    def else_stmt(self, p: YaccProduction) -> YaccProduction:
        """Else statement rule."""
        return p

    @_(
        "KW_TRY code_block",
        "KW_TRY code_block except_list",
        "KW_TRY code_block finally_stmt",
        "KW_TRY code_block except_list finally_stmt",
    )
    def try_stmt(self, p: YaccProduction) -> YaccProduction:
        """Try statement rule."""
        return p

    @_(
        "except_def",
        "except_list except_def",
    )
    def except_list(self, p: YaccProduction) -> YaccProduction:
        """Except statement list rule."""
        return p

    @_(
        "KW_EXCEPT expression code_block",
        "KW_EXCEPT expression KW_AS NAME code_block",
    )
    def except_def(self, p: YaccProduction) -> YaccProduction:
        """Except definition rule."""
        return p

    @_(
        "KW_FINALLY code_block",
    )
    def finally_stmt(self, p: YaccProduction) -> YaccProduction:
        """Except finally statement rule."""
        return p

    @_(
        "KW_FOR assignment KW_TO expression KW_BY expression code_block",
        "KW_FOR name_list KW_IN expression code_block",
    )
    def for_stmt(self, p: YaccProduction) -> YaccProduction:
        """For statement rule."""
        return p

    @_(
        "NAME",
        "name_list COMMA NAME",
    )
    def name_list(self, p: YaccProduction) -> YaccProduction:
        """List of names."""
        return p

    @_("KW_WHILE expression code_block")
    def while_stmt(self, p: YaccProduction) -> YaccProduction:
        """While statement rule."""
        return p

    @_("KW_WITH expr_as_list code_block")
    def with_stmt(self, p: YaccProduction) -> YaccProduction:
        """With statement rule."""
        return p

    @_(
        "expression",
        "expression KW_AS NAME",
        "expr_as_list COMMA NAME",
        "expr_as_list COMMA expression KW_AS NAME",
    )
    def expr_as_list(self, p: YaccProduction) -> YaccProduction:
        """Name as list rule."""
        return p

    @_(
        "KW_RAISE",
        "KW_RAISE expression",
    )
    def raise_stmt(self, p: YaccProduction) -> YaccProduction:
        """Raise statement rule."""
        return p

    @_(
        "KW_ASSERT expression",
        "KW_ASSERT expression COMMA expression",
    )
    def assert_stmt(self, p: YaccProduction) -> YaccProduction:
        """Assert statement rule."""
        return p

    @_(
        "KW_CONTINUE",
        "KW_BREAK",
        "KW_SKIP",
    )
    def ctrl_stmt(self, p: YaccProduction) -> YaccProduction:
        """Control statement rule."""
        return p

    @_("KW_DELETE expression")
    def delete_stmt(self, p: YaccProduction) -> YaccProduction:
        """Delete statement rule."""
        return p

    @_(
        "KW_REPORT expression",
    )
    def report_stmt(self, p: YaccProduction) -> YaccProduction:
        """Report statement rule."""
        return p

    @_(
        "KW_RETURN",
        "KW_RETURN expression",
    )
    def return_stmt(self, p: YaccProduction) -> YaccProduction:
        """Report statement rule."""
        return p

    @_(
        "KW_YIELD",
        "KW_YIELD expression",
    )
    def yield_stmt(self, p: YaccProduction) -> YaccProduction:
        """Yield statement rule."""
        return p

    @_(
        "ignore_stmt SEMI",
        "visit_stmt",
        "revisit_stmt",
        "disengage_stmt SEMI",
    )
    def walker_stmt(self, p: YaccProduction) -> YaccProduction:
        """Walker statement rule."""
        return p

    @_("KW_IGNORE expression")
    def ignore_stmt(self, p: YaccProduction) -> YaccProduction:
        """Ignore statement rule."""
        return p

    @_(
        "KW_VISIT expression SEMI",
        "KW_VISIT sub_name_dotted expression SEMI",
        "KW_VISIT expression else_stmt",
        "KW_VISIT sub_name_dotted expression else_stmt",
    )
    def visit_stmt(self, p: YaccProduction) -> YaccProduction:
        """Visit statement rule."""
        return p

    @_(
        "KW_REVISIT SEMI",
        "KW_REVISIT expression SEMI",
        "KW_REVISIT else_stmt",
        "KW_REVISIT expression else_stmt",
    )
    def revisit_stmt(self, p: YaccProduction) -> YaccProduction:
        """Visit statement rule."""
        return p

    @_("KW_DISENGAGE")
    def disengage_stmt(self, p: YaccProduction) -> YaccProduction:
        """Disengage statement rule."""
        return p

    @_("KW_AWAIT expression")
    def await_stmt(self, p: YaccProduction) -> YaccProduction:
        """Sync statement rule."""
        return p

    # Expression rules (precedence built into grammar)
    # ------------------------------------------------
    @_(
        "atom EQ expression",
        "KW_FREEZE atom EQ expression",
    )
    def assignment(self, p: YaccProduction) -> YaccProduction:
        """Rule for assignment statement."""
        return p

    @_("KW_HAS assignment_list SEMI")
    def static_assignment(self, p: YaccProduction) -> YaccProduction:
        """Rule for static assignment statement."""
        return p

    @_(
        "pipe",
        "pipe KW_IF expression KW_ELSE expression",
    )
    def expression(self, p: YaccProduction) -> YaccProduction:
        """Expression rule."""
        return p

    @_(
        "pipe_back",
        "pipe_back PIPE_FWD pipe",  # casting achieved here
    )
    def pipe(self, p: YaccProduction) -> YaccProduction:
        """Pipe forward rule."""
        return p

    @_(
        "elvis_check",
        "elvis_check PIPE_BKWD pipe_back",
    )
    def pipe_back(self, p: YaccProduction) -> YaccProduction:
        """Pipe backward rule."""
        return p

    @_(
        "bitwise_or",
        "bitwise_or ELVIS_OP elvis_check",
    )
    def elvis_check(self, p: YaccProduction) -> YaccProduction:
        """Expression rule."""
        return p

    @_(
        "bitwise_xor",
        "bitwise_xor BW_OR bitwise_or",
    )
    def bitwise_or(self, p: YaccProduction) -> YaccProduction:
        """Bitwise or rule."""
        return p

    @_(
        "bitwise_and",
        "bitwise_and BW_XOR bitwise_xor",
    )
    def bitwise_xor(self, p: YaccProduction) -> YaccProduction:
        """Bitwise xor rule."""
        return p

    @_(
        "shift",
        "shift BW_AND bitwise_and",
    )
    def bitwise_and(self, p: YaccProduction) -> YaccProduction:
        """Bitwise and rule."""
        return p

    @_(
        "logical",
        "logical LSHIFT shift",
        "logical RSHIFT shift",
    )
    def shift(self, p: YaccProduction) -> YaccProduction:
        """Shift expression rule."""
        return p

    @_(
        "compare",
        "compare KW_AND logical",
        "compare KW_OR logical",
        "NOT logical",
    )
    def logical(self, p: YaccProduction) -> YaccProduction:
        """Logical rule."""
        return p

    @_(
        "arithmetic",
        "arithmetic cmp_op compare",
    )
    def compare(self, p: YaccProduction) -> YaccProduction:
        """Compare rule."""
        return p

    @_(
        "term",
        "term PLUS arithmetic",
        "term MINUS arithmetic",
    )
    def arithmetic(self, p: YaccProduction) -> YaccProduction:
        """Arithmetic rule."""
        return p

    @_(
        "factor",
        "factor STAR_MUL term",
        "factor FLOOR_DIV term",
        "factor DIV term",
        "factor MOD term",
    )
    def term(self, p: YaccProduction) -> YaccProduction:
        """Term rule."""
        return p

    @_(
        "PLUS factor",
        "MINUS factor",
        "BW_NOT factor",
        "power",
    )
    def factor(self, p: YaccProduction) -> YaccProduction:
        """Factor rule."""
        return p

    @_(
        "connect",
        "connect STAR_POW power",
    )
    def power(self, p: YaccProduction) -> YaccProduction:
        """Power rule."""
        return p

    @_(
        "atomic_pipe disconnect_op connect",
        "atomic_pipe connect_op connect",
        "atomic_pipe",
    )
    def connect(self, p: YaccProduction) -> YaccProduction:
        """Connect rule."""
        return p

    @_(
        "atomic_pipe A_PIPE_FWD atomic_pipe_back",
        "atomic_pipe KW_SPAWN atomic_pipe_back",  # For high level readability
        "atomic_pipe_back",
    )
    def atomic_pipe(self, p: YaccProduction) -> YaccProduction:
        """Pipe forward rule."""
        return p

    @_(
        "atomic_pipe_back A_PIPE_BKWD unpack",
        "unpack",
    )
    def atomic_pipe_back(self, p: YaccProduction) -> YaccProduction:
        """Pipe backward rule."""
        return p

    @_(
        "STAR_POW atom",
        "STAR_MUL atom",
        "ref",
    )
    def unpack(self, p: YaccProduction) -> YaccProduction:
        """Unpack rule."""
        return p

    @_(
        "BW_AND walrus_assign",
        "walrus_assign",
    )
    def ref(self, p: YaccProduction) -> YaccProduction:
        """Unpack rule."""
        return p

    @_(
        "ds_call",
        "ds_call walrus_op walrus_assign",
    )
    def walrus_assign(self, p: YaccProduction) -> YaccProduction:
        """Walrus assignment rule."""
        return p

    @_(
        "KW_SPAWN atom",
        "A_PIPE_FWD atom",
        "PIPE_FWD atom",
        "atom",
    )
    def ds_call(self, p: YaccProduction) -> YaccProduction:
        """Unpack rule."""
        return p

    @_(
        "WALRUS_EQ",
        "ADD_EQ",
        "SUB_EQ",
        "MUL_EQ",
        "FLOOR_DIV_EQ",
        "DIV_EQ",
        "MOD_EQ",
        "BW_AND_EQ",
        "BW_OR_EQ",
        "BW_XOR_EQ",
        "BW_NOT_EQ",
        "LSHIFT_EQ",
        "RSHIFT_EQ",
    )
    def walrus_op(self, p: YaccProduction) -> YaccProduction:
        """Production Assignment rule."""
        return p

    @_(
        "EE",
        "LT",
        "GT",
        "LTE",
        "GTE",
        "NE",
        "KW_IN",
        "KW_NIN",
        "KW_IS",
        "KW_ISN",
    )
    def cmp_op(self, p: YaccProduction) -> YaccProduction:
        """Compare operator rule."""
        return p

    # Atom rules
    # --------------------
    @_(
        "atom_literal",
        "atom_collection",
        "LPAREN expression RPAREN",
        "atomic_chain",
        "all_refs",
        "edge_op_ref",
    )
    def atom(self, p: YaccProduction) -> YaccProduction:
        """Atom rule."""
        return p

    @_(
        "INT",
        "HEX",
        "BIN",
        "OCT",
        "FLOAT",
        "multistring",
        "BOOL",
        "NULL",
        "builtin_type",
    )
    def atom_literal(self, p: YaccProduction) -> YaccProduction:
        """Atom rule."""
        return p

    @_(
        "list_val",
        "tuple_val",
        "set_val",
        "dict_val",
        "list_compr",
        "gen_compr",
        "set_compr",
        "dict_compr",
    )
    def atom_collection(self, p: YaccProduction) -> YaccProduction:
        """Atom rule."""
        return p

    @_(
        "STRING",
        "FSTRING",
        "multistring STRING",
        "multistring FSTRING",
    )
    def multistring(self, p: YaccProduction) -> YaccProduction:
        """Multistring rule."""
        return p

    @_(
        "LSQUARE RSQUARE",
        "LSQUARE expr_list RSQUARE",
    )
    def list_val(self, p: YaccProduction) -> YaccProduction:
        """List value rule."""
        return p

    @_(
        "LPAREN RPAREN",
        "LPAREN tuple_list RPAREN",
    )
    def tuple_val(self, p: YaccProduction) -> YaccProduction:
        """Tuple value rule."""
        return p

    @_(
        "LBRACE expr_list RBRACE",
    )
    def set_val(self, p: YaccProduction) -> YaccProduction:
        """Set value rule."""
        return p

    @_(
        "expression",
        "expr_list COMMA expression",
    )
    def expr_list(self, p: YaccProduction) -> YaccProduction:
        """Expression list rule."""
        return p

    @_(
        "expression COMMA",
        "expression COMMA expr_list",
        "assignment_list",
        "expression COMMA assignment_list",
        "expression COMMA expr_list COMMA assignment_list",
    )
    def tuple_list(self, p: YaccProduction) -> YaccProduction:
        """Tuple list rule."""
        return p

    @_(
        "LBRACE RBRACE",
        "LBRACE kv_pairs RBRACE",
    )
    def dict_val(self, p: YaccProduction) -> YaccProduction:
        """Production for dictionary value rule."""
        return p

    @_("LSQUARE inner_compr RSQUARE")
    def list_compr(self, p: YaccProduction) -> YaccProduction:
        """Comprehension rule."""
        return p

    @_("LPAREN inner_compr RPAREN")
    def gen_compr(self, p: YaccProduction) -> YaccProduction:
        """Comprehension rule."""
        return p

    @_("LBRACE inner_compr RBRACE")
    def set_compr(self, p: YaccProduction) -> YaccProduction:
        """Comprehension rule."""
        return p

    @_(
        "expression KW_FOR name_list KW_IN walrus_assign",
        "expression KW_FOR name_list KW_IN walrus_assign KW_IF expression",
    )
    def inner_compr(self, p: YaccProduction) -> YaccProduction:
        """Comprehension rule."""
        return p

    @_(
        "LBRACE expression COLON expression KW_FOR name_list KW_IN walrus_assign RBRACE",
        "LBRACE expression COLON expression KW_FOR name_list KW_IN walrus_assign KW_IF expression RBRACE",
    )
    def dict_compr(self, p: YaccProduction) -> YaccProduction:
        """Comprehension rule."""
        return p

    @_(
        "expression COLON expression",
        "kv_pairs COMMA expression COLON expression",
    )
    def kv_pairs(self, p: YaccProduction) -> YaccProduction:
        """Key/value pairs rule."""
        return p

    @_(
        "atomic_chain_safe",
        "atomic_chain_unsafe",
        "atomic_call",
    )
    def atomic_chain(self, p: YaccProduction) -> YaccProduction:
        """Atom trailer rule."""
        return p

    @_(
        "atom DOT all_refs",
        "atom DOT_FWD all_refs",
        "atom DOT_BKWD all_refs",
        "atom index_slice",
        "atom arch_ref",
        "atom edge_op_ref",
        "atom filter_compr",
    )
    def atomic_chain_unsafe(self, p: YaccProduction) -> YaccProduction:
        """Atom trailer rule."""
        return p

    @_(
        "atom NULL_OK DOT all_refs",
        "atom NULL_OK DOT_FWD all_refs",
        "atom NULL_OK DOT_BKWD all_refs",
        "atom NULL_OK index_slice",
        "atom NULL_OK arch_ref",
        "atom NULL_OK edge_op_ref",
        "atom NULL_OK filter_compr",
    )
    def atomic_chain_safe(self, p: YaccProduction) -> YaccProduction:
        """Atom trailer rule."""
        return p

    @_("atom func_call_tail")
    def atomic_call(self, p: YaccProduction) -> YaccProduction:
        """Ability call rule."""
        return p

    @_(
        "LPAREN RPAREN",
        "LPAREN param_list RPAREN",
    )
    def func_call_tail(self, p: YaccProduction) -> YaccProduction:
        """Rule for function calls."""
        return p

    @_(
        "expr_list",
        "assignment_list",
        "expr_list COMMA assignment_list",
    )
    def param_list(self, p: YaccProduction) -> YaccProduction:
        """Parameter list rule."""
        return p

    @_(
        "assignment",
        "assignment_list COMMA assignment",
    )
    def assignment_list(self, p: YaccProduction) -> YaccProduction:
        """Keyword expression list rule."""
        return p

    @_(
        "LSQUARE expression RSQUARE",
        "LSQUARE expression COLON expression RSQUARE",
        "LSQUARE expression COLON RSQUARE",
        "LSQUARE COLON expression RSQUARE",
        "LSQUARE COLON RSQUARE",
    )
    def index_slice(self, p: YaccProduction) -> YaccProduction:
        """Index/slice rule."""
        return p

    # Architype reference rules
    # -------------------------
    @_(
        "strict_arch_ref",
        "ability_ref",
        "enum_ref",
        "global_ref",
    )
    def arch_ref(self, p: YaccProduction) -> YaccProduction:
        """Architype reference rule."""
        return p

    @_(
        "node_ref",
        "edge_ref",
        "walker_ref",
        "object_ref",
    )
    def strict_arch_ref(self, p: YaccProduction) -> YaccProduction:
        """Strict Architype reference rule."""
        return p

    @_("NODE_OP NAME")
    def node_ref(self, p: YaccProduction) -> YaccProduction:
        """Node reference rule."""
        return p

    @_("EDGE_OP NAME")
    def edge_ref(self, p: YaccProduction) -> YaccProduction:
        """Edge reference rule."""
        return p

    @_("WALKER_OP NAME")
    def walker_ref(self, p: YaccProduction) -> YaccProduction:
        """Walker reference rule."""
        return p

    @_("OBJECT_OP esc_name")
    def object_ref(self, p: YaccProduction) -> YaccProduction:
        """Object type reference rule."""
        return p

    @_("ENUM_OP NAME")
    def enum_ref(self, p: YaccProduction) -> YaccProduction:
        """Object type reference rule."""
        return p

    @_(
        "ABILITY_OP esc_name",
        "ABILITY_OP special_refs",  # only <init> is valid for now
    )
    def ability_ref(self, p: YaccProduction) -> YaccProduction:
        """Ability reference rule."""
        return p

    @_("GLOBAL_OP esc_name")
    def global_ref(self, p: YaccProduction) -> YaccProduction:
        """Global reference rule."""
        return p

    # Node / Edge reference and connection rules
    # ------------------------------------------
    @_(
        "edge_to",
        "edge_from",
        "edge_any",
    )
    def edge_op_ref(self, p: YaccProduction) -> YaccProduction:
        """Edge reference rule."""
        return p

    @_(
        "ARROW_R",
        "ARROW_R_p1 expression ARROW_R_p2",
        "ARROW_R_p1 expression COLON filter_compare_list ARROW_R_p2",
    )
    def edge_to(self, p: YaccProduction) -> YaccProduction:
        """Edge to rule."""
        return p

    @_(
        "ARROW_L",
        "ARROW_L_p1 expression ARROW_L_p2",
        "ARROW_L_p1 expression COLON filter_compare_list ARROW_L_p2",
    )
    def edge_from(self, p: YaccProduction) -> YaccProduction:
        """Edge from rule."""
        return p

    @_(
        "ARROW_BI",
        "ARROW_L_p1 expression ARROW_R_p2",
        "ARROW_L_p1 expression COLON filter_compare_list ARROW_R_p2",
    )
    def edge_any(self, p: YaccProduction) -> YaccProduction:
        """Edge any rule."""
        return p

    @_(
        "connect_to",
        "connect_from",
    )
    def connect_op(self, p: YaccProduction) -> YaccProduction:
        """Connect operator rule."""
        return p

    @_("NOT edge_op_ref")
    def disconnect_op(self, p: YaccProduction) -> YaccProduction:
        """Connect operator not rule."""
        return p

    @_(
        "CARROW_R",
        "CARROW_R_p1 expression CARROW_R_p2",
        "CARROW_R_p1 expression COLON assignment_list CARROW_R_p2",
    )
    def connect_to(self, p: YaccProduction) -> YaccProduction:
        """Connect to rule."""
        return p

    @_(
        "CARROW_L",
        "CARROW_L_p1 expression CARROW_L_p2",
        "CARROW_L_p1 expression COLON assignment_list CARROW_L_p2",
    )
    def connect_from(self, p: YaccProduction) -> YaccProduction:
        """Connect from rule."""
        return p

    @_("LPAREN EQ filter_compare_list RPAREN")
    def filter_compr(self, p: YaccProduction) -> YaccProduction:
        """Filter context rule."""
        return p

    @_(
        "esc_name cmp_op expression",
        "filter_compare_list COMMA esc_name cmp_op expression",
    )
    def filter_compare_list(self, p: YaccProduction) -> YaccProduction:
        """Filter comparison list rule."""
        return p

    @_("")
    def empty(self, p: YaccProduction) -> YaccProduction:
        """Empty rule."""
        return p

    # Transform Implementations
    # -------------------------
    def transform(self, ir: list) -> ast.AstNode:
        """Tokenize the input."""
        ir = self.parse(ir)
        return ir

    def error(self, p: YaccProduction) -> None:
        """Improved error handling for Jac Parser."""
        self.cur_line = p.lineno if p else 0
        if not p:
            self.log_error("Escaping at end of File! Not Valid Jac!\n")
            return
        self.log_error(f'JParse Error, incorrect usage of "{p.value}" ({p.type})\n')

        # Read ahead looking for a closing '}'
        while True:
            tok = next(self.tokens, None)  # type: ignore
            if not tok or tok.type == "RBRACE":
                break
        self.restart()


def fstr_sly_parser_hack() -> Optional[dict]:
    """Hack to map expression parser for fstrings in sly parser."""
    if "__file__" in globals():
        with open(__file__, "r") as file:
            module_data = file.read()

        new_module_data = module_data.replace(
            'start = "module"', 'start = "expression"'
        ).replace('debugfile = "parser.out"', "")
        new_module_namespace = {}
        exec(new_module_data, new_module_namespace)
        return (
            new_module_namespace["JacParser"]
            if "JacParser" in new_module_namespace
            else None
        )


JacParserExpr = fstr_sly_parser_hack()


def parse_tree_to_ast(
    tree: tuple, parent: Optional[ast.AstNode] = None, lineno: int = 0
) -> ast.AstNode:
    """Convert parser output to ast, also parses fstrings."""

    def find_and_concat_fstr_pieces(tup: tuple) -> str:
        result = ""
        for item in tup:
            if isinstance(item, tuple):
                result += find_and_concat_fstr_pieces(item)
            elif isinstance(item, LexToken) and item.type == "PIECE":
                result += item.value
        return result

    from jaclang.utils.fstring_parser import FStringLexer, FStringParser
    from jaclang.utils.sly.lex import Token as LexToken

    ast_tree: ast.AstNode = None
    if not isinstance(tree, ast.AstNode):
        if isinstance(tree, tuple):
            if tree[0] == "fstr_expr":
                tree = JacParserExpr(
                    mod_path="",
                    input_ir=JacLexer(
                        mod_path="", input_ir=find_and_concat_fstr_pieces(tree)
                    ).ir,
                ).ir_tup[2]
            kids = tree[2:]
            ast_tree = ast.Parse(
                name=tree[0],
                parent=parent,
                mod_link=None,
                line=tree[1] if lineno == 0 else lineno,
                kid=[],
            )
            ast_tree.kid = [
                parse_tree_to_ast(x, parent=ast_tree, lineno=lineno) for x in kids
            ]
        elif isinstance(tree, LexToken):
            if tree.type == "FSTRING":
                lineno = tree.lineno
                ftree = FStringParser().parse(FStringLexer().tokenize(tree.value))
                return parse_tree_to_ast(ftree, parent=parent, lineno=lineno)
            else:
                meta = {
                    "name": tree.type,
                    "parent": parent,
                    "mod_link": None,
                    "value": tree.value,
                    "kid": [],
                    "line": tree.lineno if lineno == 0 else lineno,
                    "col_start": tree.index - tree.lineidx + 1,
                    "col_end": tree.end - tree.lineidx + 1,
                }
                if tree.type == "NAME":
                    ast_tree = ast.Name(already_declared=False, **meta)
                elif tree.type == "KWESC_NAME":
                    meta["value"] = meta["value"].replace("<>", "")
                    ast_tree = ast.Name(already_declared=False, **meta)
                elif tree.type == "FLOAT":
                    ast_tree = ast.Constant(typ=float, **meta)
                elif tree.type in ["INT", "HEX", "BIN", "OCT"]:
                    ast_tree = ast.Constant(typ=int, **meta)
                elif tree.type in ["STRING", "FSTRING"]:
                    ast_tree = ast.Constant(typ=str, **meta)
                elif tree.type == "BOOL":
                    ast_tree = ast.Constant(typ=bool, **meta)
                elif tree.type == "NULL":
                    ast_tree = ast.Constant(typ=type(None), **meta)
                elif tree.type.startswith("TYP_"):
                    ast_tree = ast.Constant(typ=type, **meta)
                else:
                    ast_tree = ast.Token(**meta)
        else:
            raise ValueError("Syntax Error encountered while parsing Jac program.")
    if not ast_tree:
        raise ValueError(f"node must be AstNode: {tree}")
    return ast_tree
